# include <math.h>
# include <stdio.h>
# include <stdlib.h>

# include <hstio.h>     /* defines HST I/O functions */
# include "calnic.h"	/* defines NICMOS data structures */
# include "calnica.h"	/* defines CALNICA data structures */

# define THRESH 4.0		/* default sigma threshold for rejection */

# define DEBUG          0
# define X1		167
# define Y1		119
# define X2		1       /* not used*/
# define Y2		1       /*not used*/


# define max_CRs         3
# define equal_weight    0
# define optimum_weight  1
# define clean_up        1
# define calculate       0
# define DQIGNORE        SATURATED

static void fitsamps (short, float *, float *, short *, float *, float *,
                      float, float *, float *, short *, float *, short,
                      short, float, float);

static void linfit (short, float *, float *, float *, 
                    float *, short, 
                    float, float, 
                    float *, float *, float *, float *);

static int RejSpikes (float *, float *, short *, float *, short, float, int *);
static int RejFirstRead (short *, float *, short, float);
static int RejCRs (short *, float *, short, float, int *);
static void GetDarkAndGlow(short operation, char dark_ref_name[SZ_NAME+1],
                           NicInfo *nic, short first_read, short last_read, 
			   float *time, short i, short j,
                           float *Poisson_counts);
static void EstimateDarkandGlow(short operation, NicInfo *nic, short nsamp,
                                short row, short col, float * time,
                                float *tot_ADUs);
/* N_CRIDCALC: Identify cosmic ray hits in NICMOS images. This routine
** calls separate subroutines for working with single (ACCUM,RAMP,BRTOBJ)
** images or multiple (MULTIACCUM) images.
**
** Revision history:
** H.Bushouse	Feb. 1996	Written for Build 2 (Version 2.0)
** H.Bushouse	05-Jun-1997	New n_mcrid routine that performs a linear
**				fit to flux vs time and uses NOISFILE to
**				compute errors. (Version 2.3)
** H.Bushouse	23-Jul-1997	Changed linfit to return zero when given 0
**				data points and return the input value when
**				given 1 data point. (Version 3.0)
** H.Bushouse	28-Jul-1997	Changed NicInfo.bunit from scalar to vector
**				(Version 3.0)
** H.Bushouse	11-Sep-1997	Added use of nic.crthresh to allow user input
**				(Version 3.0)
** H.Bushouse	29-Sep-1997	Complete rewrite of n_mcrid - eliminated use of
**				asigclip,aerrclip,merrclip,mean,wtmean routines
**				to compute mean differences and reject samples.
**				Rejection and fit to samples are now both done
**				together using new fitsamps routine. No need
**				for NOISFILE anymore. (Version 3.0)
** H.Bushouse	15-Feb-1998	Modified n_mcrid and fitsamps for special
**				handling of MultiAccum zeroth reads when
**				first read is saturated, including special
**				handling of new	ZEROSIG and GROT DQ values
**				(Version 3.2)
** H.Bushouse	05-May-1998	Modified linfit to fix division by zero bug
**				when ngood=1 and x[goodpt]=0 (Version 3.2)
** H.Bushouse	30-Sep-1998	Modified n_mcrid to use TIME image data for
**				doing count<->countrate conversions, instead of
**				using nic->exptime, since individual pixels may
**				have different exposure times (Version 3.3)
** H.Bushouse	01-Oct-1998	Modified n_mcrid to flag initial samples based
**				on value of nic->samp_rej (Version 3.3)
** H.Bushouse	05-Oct-1998	Put sections of code in fitsamps that flag
**				electronic noise spikes and CR hits into their
**				own functions "RejSpikes" and "RejCRs".
**				Modified RejSpikes so that only the largest
**				outlier is rejected on each pass. Modified
**				fitsamps to reiterate the fit after rejecting
**				spikes before looking for for CR hits (Vsn 3.3)
** H.Bushouse	16-Sep-1999	Moved static function prototypes to head of
**				file (to satisfy DEC/Alpha/OpenVMS compiler)
**				(Version 3.3)
** H.Bushouse	16-Jun-2000	Added check for ZSIG.done=PERFORMED in n_mcrid
**				routine, in case it was done in a previous run.
**				Moved call to n_statcalc out of this module
**				and into n_docalib2, so that it gets performed
**				after all other possible calibration steps.
**				(Version 4.0)
** M. Regan     07-May-2007     Major revision:  
                                1. Changed implementation of how the slope is 
				   determined when CRs are found. A weighted
                                   mean of the fits in each CR-free interval
                                   is now returned.
				2. Now uses the optimum weighting method of
                                   Fixsen et al. for the slope fits at high
                                   S/N. At low S/N changed to uniform
                                   weighting rather than including the
                                   Poisson uncertainty.
			        3. The returned error plane now includes
                                   Poisson noise, flat field error,
                                   uncertainty in the fit, and Poisson from
                                   the dark current.
				4. Changed the calculation of the returned
                                   time plane to reflect the net integration
                                   time after CRs and zero read subtraction.
				5. Changed the method of indentifying
                                   electronic spikes to better find them.

V. Laidler   12 Jun 07          Bugfixes & refactoring to fitsamps, which
                                will now only pass in data that should
                                actually be fit to linfit.
			  
V. Laidler   14 Jun 07          Bugfix & significant refactoring to linfit:
                                argument list has changed as mask is no
                                longer passed in. Contributions to error now
                                includes flat and dark&glow for 2-point cases.

V. Laidler   22 Jun 07          Significant refactoring to fitsamps;
                                intervals are now stored separately & tests
                                are done in only one place. Calling sequences
                                of RejSpikes and RejCRs also changed.

V. Laidler   26 Jun 07          Temporarily added some debug print statements
                                that can be enabled for a particular pixel.
V. Laidler   9 Jul 07           Only check first read of interval for spikes
                                after everything else has already been
                                identified.

R. Jedrzejewski  26 Oct 07      Set pixels with more than max_CRs to a sensible
                                value.  Flag with new HIGH_CURVATURE bit
                                (4096).  Set max_CRs to 4.  Unset CR_HIT bit
                                if HIGH_CURVATURE bit set.
R. Jedrzejewski  05 Nov 07      Sensible value for HIGH_CURVATURE pixels =
                                (last good counts - first good counts)/(last
                                good time - first good time).  Count number of
                                HIGH_CURVATURE pixels instead of printing out
R. Jedrzejewski  30 Nov 07      Change max_CRs in n_cridcalc.c from 4 to 3
R. Jedrzejewski  05 Mar 08      Do up-the-ramp processing for pixels that
                                don't have any DQ=0 reads.  Propagate DQ
                                flags to the _cal file.  Version 4.2
R. Jedrzejewski  18 Apr 08      Restore 4096 as USER1 DQ flag.  Use 16384
                                (previously RESERVED1) for HIGH_CURVATURE
R. Jedrzejewski  01 Aug 08      Bug fix, error should always be calculated
                                in linfit, not just when dy > 0.0

M. Sosey		Oct 30 2008     Updated to account for temperature dependent
								ampglow and linear dark images in the TDD


*/

int n_cridcalc (NicInfo *nic, MultiNicmosGroup *input,
		SingleNicmosGroup *crimage, SingleNicmosGroup *flat) {

  /* Arguments:
**	nic	 i: NICMOS info structure
**	input	io: input image(s)
**	crimage	 o: CR reject output image (for MultiACCUM only)
*/

	/* Function definitions */
  int n_mcrid (NicInfo *, MultiNicmosGroup *, SingleNicmosGroup *,
               SingleNicmosGroup *);
  int n_scrid (NicInfo *, MultiNicmosGroup *);
  int updatePriHdr (CalStep *, Hdr *);
  int updateSciHdr (CalStep *, Hdr *);
  int n_calReport (CalStep *, int, Hdr *, Hdr *);

  if (nic->CRID.corr == PERFORM) {

    /* Call multi-image routine for MultiACCUM datasets */
    if (nic->obsmode == MULTIACCUM) {
      if (n_mcrid (nic, input, crimage, flat))
	return (status);

      if (updatePriHdr (&nic->CRID, crimage->globalhdr))
	return (status);
      if (updateSciHdr (&nic->CRID, &(crimage->sci.hdr)))
	return (status);

      /* Call single-image routine for all others */
    } else {
      if (n_scrid (nic, input))
	return (status);
    }
  }

  for (nic->group=nic->ngroups; nic->group >= 1; nic->group--) {
    n_calReport (&nic->CRID, nic->group,
		 &input->group[nic->group-1].sci.hdr,
		 input->group[nic->group-1].globalhdr);
  }

  /* Successful return */
  return (status = 0);
}

/* N_SCRID: Identify cosmic ray hits in single NICMOS images.
** Detected hits are only flagged in the DQ array; SCI image pixel
** values are NOT changed. */

int n_scrid (NicInfo *nic, MultiNicmosGroup *input) {

  /* Arguments:
  **	nic	 i: NICMOS info structure
  **	input	 i: input image
  */

  /* Routine not available yet */
  sprintf (MsgText, "CRIDCALC not yet implemented; will be skipped");
  n_warn (MsgText);

  nic->CRID.corr = OMIT;

  /* Successful return */
  return (status = 0);
}

/* N_MCRID: Identify cosmic ray hits in a stack of MultiACCUM samples.
** Detected hits are flagged in the DQ arrays of the individual samples,
** but the SCI images in the individual samples are NOT changed.
** A final single-group image is produced using only unflagged data
** from the individual samples. The accumlated counts vs. time are fitted
** with a first-order (linear) polynomial using standard linear regression.
** Outliers are identified by their distance from the fit and rejected. */

int n_mcrid (NicInfo *nic, MultiNicmosGroup *input, SingleNicmosGroup *crimage,
             SingleNicmosGroup *flat){

  /* Arguments:
  **	nic	 i: NICMOS info structure
  **	input	io: input image stack
  **	crimage	 o: CR reject output image (single-group)
  */

  /* Local variables */
  short i, j, k, l, jj;		/* pixel and loop indexes */
  short nsamp;			/* number of samples for pixel */
  float current_sci;		/* current sci value */
  float current_err;		/* current err value */
  short current_dq;		/* current dq  value */
  float *sci;			/* list of sci  values for pixel */
  float *err;			/* list of err  values for pixel */
  short *dq;			/* list of dq   values for pixel */
  float *time;			/* list of time values for pixel */
  float *tot_ADUs;		/* list of total ADUs from dark current and 
                                   amp glow for pixel */
  float out_sci;		/* output sci  value */
  float out_err;		/* output err  value */
  short out_dq;			/* output dq   value */
  short out_samp;		/* output samp value */
  float out_time;		/* output time value */
  float flat_value;             /* value to convert from flat fielded ADUs to 
                                   electrons */
  float flat_uncertainty;       /* unitless rms flat field uncertainty */
  float poisson_counts;         /* total ADUs for this sample from dark
                                   current and amp glow */
  int ncurved;			/* Number of pixels with high curvature */

  /* Function definitions */
  int n_copyGroup (SingleNicmosGroup *, SingleNicmosGroup *);

  /* Allocate memory for local arrays */
  sci    = NULL;
  err    = NULL;
  dq     = NULL;
  time   = NULL;
  sci    = (float *) calloc(nic->ngroups, sizeof(float));
  err    = (float *) calloc(nic->ngroups, sizeof(float));
  dq     = (short *) calloc(nic->ngroups, sizeof(short));
  time   = (float *) calloc(nic->ngroups, sizeof(float));
  tot_ADUs   = (float *) calloc(nic->ngroups, sizeof(float));

  /* Initialize the output crimage by copying the first input
  ** group to the crimage */
  if (n_copyGroup (crimage, &(input->group[0])))
    return (status);

  /* Set CR reject threshold; use default if no user input */
  if (nic->crthresh == 0)
    nic->crthresh = THRESH;

  /* Print info to processing log */
  if (nic->samp_rej > 0) {
    sprintf (MsgText, "CRIDCALC using %g sigma rejection threshold",
	     nic->crthresh);
    n_message (MsgText);
    sprintf (MsgText,
	     "         and rejecting first %d samples for all pixels",
	     nic->samp_rej);
    n_message (MsgText);
  } else {
    sprintf (MsgText, "CRIDCALC using %g sigma rejection threshold",
	     nic->crthresh);
    n_message (MsgText);
  }

  ncurved = 0;

  /* Loop over image array, computing mean countrate at each pixel */
  for (j=0; j<input->group[0].sci.data.ny; j++) {
    for (i=0; i<input->group[0].sci.data.nx; i++) {

      /* Get the DQ value in the zeroth-read */
      k = nic->ngroups-1;
      current_dq = DQPix(input->group[k].dq.data,i,j);

      /* Back out the ZEROSIG and GROT dq flags */
      if (current_dq & ZEROSIG)
	current_dq -= ZEROSIG;
      if (current_dq & GROT)
	current_dq -= GROT;

      /* Check for pixels that are saturated already in first read
      ** but have OK data in the zeroth read */
      if ((nic->ZSIG.corr == PERFORM || nic->ZSIG.done == PERFORMED) &&
	  DQPix(input->group[k-1].dq.data,i,j) & SATURATED &&
	  current_dq == 0) {

	/* For these pixels, just set the output values equal
	** to what's in the input zeroth-read image */
	Pix(crimage->sci.data,i,j) = Pix(input->group[k].sci.data,i,j);
	Pix(crimage->err.data,i,j) = Pix(input->group[k].err.data,i,j);
	DQSetPix(crimage->dq.data,i,j,
		 DQPix(input->group[k].dq.data,i,j));
	Pix(crimage->smpl.data,i,j) = 1;
	Pix(crimage->intg.data,i,j) = nic->sampzero;

	continue;
      }

      /* Initialize the output image DQ value */
      out_dq = -1;

      /* Create list of samples for this pixel */
      nsamp = 0;
      for (k = nic->ngroups-1; k >= 0; k--) {

	/* Retrieve SCI, ERR, and DQ values for this sample */
	current_sci = Pix(input->group[k].sci.data,i,j);
	current_err = Pix(input->group[k].err.data,i,j);
	current_dq  = DQPix(input->group[k].dq.data,i,j);

	/* Temporarily convert countrates back to counts */
	if (nic->bunit[0] == COUNTRATE) {
	  current_sci *= Pix(input->group[k].intg.data,i,j);
	  current_err *= Pix(input->group[k].intg.data,i,j);
	}

	/* Propagate DQ values to output DQ if they appear in every
           sample */
	out_dq = out_dq & current_dq;

	/* Remove ZEROSIG and GROT bits from DQ, if present,
	** because those samples are OK to use here (Vsn 3.2) */
	if (current_dq & ZEROSIG)
	  current_dq -= ZEROSIG;
	if (current_dq & GROT)
	  current_dq -= GROT;

	/* Temporarily flag first samp_rej samples to exclude
	** them from the fit (Version 3.3) */
	if (nic->samp_rej > 0 && nsamp > 0 && nsamp <= nic->samp_rej)
	  current_dq = current_dq | RESERVED2;

	/* Copy all values to the fitting arrays */
	sci[nsamp]  = current_sci;
	err[nsamp]  = current_err;
	dq[nsamp]   = current_dq;
	time[nsamp] = Pix(input->group[k].intg.data,i,j);
	nsamp++;
      }
      if (nic->DARK.corr == PERFORM){
	if (strcmp (nic->darkmeth, "TEMPERATURE-DEPENDENT") == 0) {
	  GetDarkAndGlow(calculate, nic->DARK.ref.name, nic, 0, nsamp-1, time,
                         i, j, tot_ADUs);
	}else{
	  EstimateDarkandGlow(calculate, nic, nsamp, i, j, time, tot_ADUs);
	}
      }/*tot_ADUs already init'd to zero; thus no "else" clause necessary*/
	  

      if (nic ->FLAT.corr != PERFORM) {
	flat_value = (nic->adcgain);
	flat_uncertainty = 0.0;
      } else {
	if (Pix(flat->sci.data,i,j) > 0.0) {
	  flat_value = (nic->adcgain)/Pix(flat->sci.data,i,j);
	  flat_uncertainty = Pix(flat->err.data,i,j);
	} else {
	  flat_value = (nic->adcgain);
	  flat_uncertainty = 0.0;
	}
      }
      /* Do iterative rejection and computation of slope */
      fitsamps (nsamp, sci, err, dq, time, tot_ADUs, nic->crthresh,
		&out_sci, &out_err, &out_samp, &out_time, i, j, flat_value,
                flat_uncertainty);

      /* Propagate all DQ flags to output EXCEPT for SATURATED
      ** and CR_HIT  (Version 4.2) */

      out_dq = out_dq - (out_dq & SATURATED);

      /* Convert results back to counts if necessary */
      if (nic->bunit[0] == COUNTS) {
	out_sci *= out_time;
	out_err *= out_time;
      }

      /* Set CR_HIT DQ for all samples following a hit */
      for (k = 0; k < nic->ngroups-1; k++) {
	if (dq[k] & CR_HIT) {
	    dq[k+1] = dq[k+1] | CR_HIT;
        }
      }

      /* If the HIGH-CURVATURE bit is set anywhere, set it for all
         groups and unset the CR_HIT bit */
      for (k = 0; k < nic->ngroups-1; k++) {
        if (dq[k] & HIGH_CURVATURE) {
          for (l=0; l < nic->ngroups; l++) {
            dq[l] = dq[l] | HIGH_CURVATURE;
            dq[l] = dq[l] - (dq[l] & CR_HIT);
          }
          break;
        }
      }

      /* Add HIGH_CURVATURE bit to output crimage if set in the dq extension
      ** and unset the CR_HIT bit  */ 
      if (dq[nsamp-1] & HIGH_CURVATURE) {
        out_dq += HIGH_CURVATURE;
        out_dq = out_dq - (out_dq & CR_HIT);
	/* And increment the ncurved counter */
	ncurved++;
      }  /*

      /* If no points were fit, they were all saturated */
      if (out_err >= 99999.0) {
	out_dq = out_dq | SATURATED;
      }
      /* Store final values in output crimage */
      Pix(crimage->sci.data,i,j)  = out_sci;
      Pix(crimage->err.data,i,j)  = out_err;
      DQSetPix(crimage->dq.data,i,j,out_dq);
      Pix(crimage->smpl.data,i,j) = out_samp;
      Pix(crimage->intg.data,i,j) = out_time;

      /* Update input DQ values for detected outliers */
      for (k = nic->ngroups-1; k >= 0; k--) {

	if (dq[nic->ngroups-1-k] & CR_HIT)
	  DQSetPix (input->group[k].dq.data,i,j,
		    DQPix (input->group[k].dq.data,i,j) | CR_HIT);

	if (dq[nic->ngroups-1-k] & BADPIX)
	  DQSetPix (input->group[k].dq.data,i,j,
		    DQPix (input->group[k].dq.data,i,j) | BADPIX);

        if (dq[nic->ngroups-1-k] & HIGH_CURVATURE)
          DQSetPix (input->group[k].dq.data,i,j,
                    DQPix (input->group[k].dq.data,i,j) | HIGH_CURVATURE);

      }

    } /* end of loop over nx */
  } /* end of loop over ny */

  if (nic->DARK.corr == PERFORM) {
    if (strcmp (nic->darkmeth, "TEMPERATURE-DEPENDENT") == 0) {
      GetDarkAndGlow(clean_up, nic->DARK.ref.name, nic, 0, nsamp-1, time, i,
                     j, tot_ADUs);
    } else {
      EstimateDarkandGlow(clean_up, nic, nsamp, i, j, time, tot_ADUs);
    }
  } /*No cleanup necessary if they were never called*/

  if (ncurved > 0) {
    sprintf (MsgText, "%d pixels detected with high curvature", ncurved);
    n_message (MsgText);
  }

  /* Free memory allocated locally */
  free (sci);
  free (err);
  free (dq);
  free (time);

  /* Successful return */
  return (status = 0);
}


/* FITSAMPS: Fit accumulating counts vs. time to compute mean countrate,
** iteratively rejecting CR hits and refitting until no new samples are
** rejected. The rejection test is based on the input error value for
** each sample.
*/

static void fitsamps (short nsamp, float *sci, float *err, short *dq,
		      float *time, float *darkandglow, float thresh, 
		      float *out_sci, float *out_err, 
		      short *out_samp, float *out_time,
		      short i, short j, float gain, float flat_uncertainty) {
  /* Local variables */
  int k,idx,l;			/* loop index */
  int nrej, rej;		/* number & index of rejected samples */
  float (*tsci)[nsamp];		/* list of cleaned fluxes */
  float (*terr)[nsamp];		/* list of cleaned errors */
  float (*ttime)[nsamp];	/* list of cleaned times */
  short (*tdq)[nsamp];		/* list of cleaned dqa */
  float (*tdarkandglow)[nsamp]; /* poisson counts for this interval */
  int   (*tlookup)[nsamp];      /* poisson counts for this interval */
  short   interval;             /* number of intervals that have to be fit */
  int   sample;
  int   number_CRs;             /* total number of CRs found in the set of
                                   reads */
  float a[max_CRs+1], siga[max_CRs+1];	/* linear fit zeropoint and error */
  float b[max_CRs+1], sigb[max_CRs+1];	/* linear fit slope and error */
  short tcount[max_CRs+1];        /* length in each interval */
  float *diff;		        /* sample differences relative to fit */
  float sumwts,sum;             /* used to combine the slopes of the CR
                                   created intervals */
  float int_time;               /* the sum of all the times without CR hits */
  float sigsquared;
  short iter_count;             /* for debugging*/
  int found_error;              /* Flag for too many cosmic rays */
  int no_samples;		/* Flag for no valid samples */
  int firstgood;                /* First pixel with OK DQ */
  int lastgood;                 /* Last pixel with OK DQ */
  int ngood;
  int nsflag;                   /* Flag set when we go through no_samples=1 */
  int naccum;

  /* Initialize the output values */
  *out_sci  = 0;
  *out_err  = 0;
  *out_samp = 0;
  *out_time = 0;

  iter_count = 0;
  found_error = 0;
  no_samples = 0;
  nsflag = 0;

  /* Allocate memory for local arrays */
  tsci   = calloc(nsamp*(max_CRs+1), sizeof(float));
  terr   = calloc(nsamp*(max_CRs+1), sizeof(float));
  tdq   =  calloc(nsamp*(max_CRs+1), sizeof(short));
  ttime  = calloc(nsamp*(max_CRs+1), sizeof(float));
  tdarkandglow   = calloc(nsamp*(max_CRs+1), sizeof(float));
  tlookup   = calloc(nsamp*(max_CRs+1), sizeof(int));
  diff   = (float *) calloc(nsamp, sizeof(float));


  /* Fit and test samples for this pixel until no new samples
  ** are rejected:

     start_index and end_index are paired variables defining the start
     and end of intervals along the samples. If no CRs were found, then 
     start_index[0] would be 1 (skip the 0th read) and
     end_index[0] would be nsamp-1.

     Finding a CR ends one interval and begins the next. The CR must
     be included in the next interval.

     This is an iterative process:
        - first, intervals are defined based on existing CRs
	- then, each interval is fitted separately
	- then, each interval is inspected for spikes
	- then, each interval is inspected for CRs

     If any spikes or CRs are found, the entire procedure repeats: new
     intervals are defined, etc.

  After the iteration ends because no new spikes or CRs are found,
  each interval is fitted separately once again, with optimum weighting,
  and then the results for each interval are combined to obtain the
  final solution for the pixel.

  */
  number_CRs=0;
  nrej = 1; 
  while (nrej > 0) {
    nrej = 0;  /* RejSpikes and RejCRs may increment nrej */
    iter_count++;

    /*---------------------------------------------------------------
             IDENTIFY THE INTERVALS
     ----------------------------------------------------------------*/
    interval = 0;
    sample = 0;
    for (k=1; k <nsamp; k++) {
      if (dq[k] == 0) { /*accumulate the sample*/
	tsci[interval][sample]     =     sci[k];
	ttime[interval][sample]    =     time[k];
	terr[interval][sample]     =     err[k];
	tdq[interval][sample]      =     dq[k];
	tdarkandglow[interval][sample] = darkandglow[k];
	tlookup[interval][sample]  =     k;
	sample++;
      } else if (dq[k] == CR_HIT) { /* When a CR is found, */
	tcount[interval]=sample; /*save the num. of points in this interval*/
	interval++;               /*start a new interval*/
	sample=0;                 /*resetting its sample count*/
	tsci[interval][sample]     =     sci[k];
	ttime[interval][sample]    =     time[k];
	terr[interval][sample]     =     err[k];
	tdq[interval][sample]      =     dq[k];
	tdarkandglow[interval][sample] = darkandglow[k];
	tlookup[interval][sample]  =     k;
	sample++;
      }
    }

    /*If the last sample we kept has a CR, then the last interval
      has nothing in it but one CR: it can't be used, so ignore it*/
    if (sample >= 1) {
      if (tdq[interval][sample-1] == CR_HIT) {
	interval--;
      } else {  /*Otherwise close the last interval*/
	tcount[interval]=sample;
      }
    } else {
      /*There are no valid samples in this pixel. Bail out.*/
      no_samples = 1;
      nrej = 0;
      break;
    }

    /******  Print statements that can be turned on for debugging *****/
    if (DEBUG) {
      /* Set the pixel you want to examine up top */

      if ((i==X1) && (j==Y1)){
	printf("========== Debug info: iteration %d ====================\n",
               iter_count);
	printf("Pixel i,j=%d,%d:    nsamp %d\n",i,j,nsamp);
	printf("Sci: ");
	for (k=0; k<nsamp; k++){
	  printf("%  10.6f",sci[k]);
	}
	printf("\nDQ: ");
	for (k=0; k<nsamp; k++){
	  printf("%  5d",dq[k]);
	}
	printf("\nTIME: ");
	for (k=0; k<nsamp; k++){
	  printf("%  10.6f",time[k]);
	}

	printf("\n\n Intervals: %d  \n", interval+1);
	for (idx=0; idx<=interval; idx++){
	  printf("Interval %d, tcount %d\n sci: ",idx+1,tcount[idx]);
	  for (k=0; k<tcount[idx]; k++){
	    printf("%  10.6f",tsci[idx][k]);
	  }
	  printf("\n  dq:");
	  for (k=0; k<tcount[idx]; k++){
	    printf("%  5d",tdq[idx][k]);
	  }
	  printf("\ntime:");
	  for (k=0; k<tcount[idx]; k++){
	    printf("%  10.6f",ttime[idx][k]);
	  }
	  printf("\n...........................\n");
	}  /* End for loop over idx  */
      }    /* End if DEBUG pixel */
    }      /* End if DEBUG

    /*****************end debugging***********************/
    

    /*---------------------------------------------------------------
             FIT EACH (non-zero-length) INTERVAL SEPARATELY
     ----------------------------------------------------------------*/

    /* We need a unique solution for each interval. */
    for (idx=0;(idx<=interval)&&nrej==0;idx++){
      if (tcount[idx] > 1){
	/* Compute mean countrate using linear fit and equal weighting 
	   to best find spikes and CRs*/
	linfit (equal_weight, ttime[idx], tsci[idx], terr[idx], 
		tdarkandglow[idx], tcount[idx], gain, flat_uncertainty, 
		&a[idx], &b[idx], &siga[idx], &sigb[idx]);
	      
    /*---------------------------------------------------------------
             EXAMINE EACH INTERVAL FOR SPIKES AND CRS
     ----------------------------------------------------------------*/

	/* Compute difference of each sample from the local fit */
	for (k = 0; k <tcount[idx]; k++) {
	  if (terr[idx][k] != 0 && tdq[idx][k] == 0){
	    diff[k] =(tsci[idx][k]-(a[idx]+b[idx]*ttime[idx][k]))
                          / terr[idx][k];
	  }
	  else{ /*but ignore diffs for already-identified CRs & spikes*/
	    diff[k] = 0;
	  }
	}

	/* Look for and flag new electronic noise spikes and CRs*/
	nrej=RejSpikes (tsci[idx], terr[idx], tdq[idx], diff, tcount[idx],
                   thresh, &rej);
	if (nrej > 0) {
	  dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] | BADPIX;
	} else if (nrej == 0) {
	/* Only look for new CR hits if no spike was found */
	  nrej = RejCRs (tdq[idx], diff, tcount[idx], thresh, &rej);
	  if (nrej > 0) {
	    dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] | CR_HIT;
	    number_CRs +=1;
	    if (number_CRs >= max_CRs) {
	      found_error = 1;
              dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] - CR_HIT;
              dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] | HIGH_CURVATURE;
	      nrej = 0; /*To break out of the while loop*/
	      break;
	    }
	  } else if (nrej == 0) {
	/* Only check the first read if nothing else has been found */
	    nrej = RejFirstRead(tdq[idx], diff, tcount[idx], thresh);
	    if (nrej > 0){
	      dq[tlookup[idx][0]] = dq[tlookup[idx][0]] | BADPIX;
	    }
	  }
	}
      } else if (tcount[idx] == 1) {
	if (tdq[idx][0] != CR_HIT) { /*ignore CR hits if they're alone*/
	  b[idx]=tsci[idx][0]/ttime[idx][0];
	  sigb[idx]=terr[idx][0]/ttime[idx][0];
	}
      }   /*endif tcount for this interval*/
    }     /*for each interval*/
  }       /*while new rejections occur*/

  /*--------------------------------------------------------------------
           If there are no good pixels, repeat this process for all
           pixels that don't have bits in the DQIGNORE pattern set
    --------------------------------------------------------------------*/

  if (no_samples == 1) {
    no_samples = 0;      /* Reset because we're starting again   */
    nsflag = 1;
    if (DEBUG) printf ("no_samples = 1 at i = %d, j = %d\n", i, j);
    number_CRs=0;
    nrej = 1;
    iter_count = 0;
    while (nrej > 0) {
      nrej = 0;  /* RejSpikes and RejCRs may increment nrej */
      iter_count++;

      /*---------------------------------------------------------------
	IDENTIFY THE INTERVALS
	----------------------------------------------------------------*/
      interval = 0;
      sample = 0;
      for (k=1; k <nsamp; k++) {
	if ((dq[k] & (DQIGNORE | CR_HIT)) == 0) { /*accumulate the sample*/
	  tsci[interval][sample]     =     sci[k];
	  ttime[interval][sample]    =     time[k];
	  terr[interval][sample]     =     err[k];
	  tdq[interval][sample]      =     dq[k];
	  tdarkandglow[interval][sample] = darkandglow[k];
	  tlookup[interval][sample]  =     k;
	  sample++;
	} else if ((dq[k] & CR_HIT) != 0) { /* When a CR is found, */
	  tcount[interval]=sample; /*save the num. of points in this interval*/
	  interval++;               /*start a new interval*/
	  sample=0;                 /*resetting its sample count*/
	  tsci[interval][sample]     =     sci[k];
	  ttime[interval][sample]    =     time[k];
	  terr[interval][sample]     =     err[k];
	  tdq[interval][sample]      =     dq[k];
	  tdarkandglow[interval][sample] = darkandglow[k];
	  tlookup[interval][sample]  =     k;
	  sample++;
	}
      }

      /*If the last sample we kept has a CR, then the last interval
	has nothing in it but one CR: it can't be used, so ignore it*/
      if (sample >= 1) {
	if (tdq[interval][sample-1] == CR_HIT) {
	  interval--;
	} else {  /*Otherwise close the last interval*/
	  tcount[interval]=sample;
	}
      } else {
	/*There are no valid samples in this pixel. Bail out.*/
	no_samples = 1;
	nrej = 0;
        if (DEBUG) printf ("Still no valid samples\n");
	break;
      }
      if (DEBUG) {
        printf ("Iter count = %d. %d intervals\n", iter_count, interval+1);
	for (k=0; k <= interval; k++) {
	  for (l=0; l < tcount[k]; l++) {
	    printf (" %d", tdq[k][l]);
	  }
	  printf ("\n");
	  for (l=0; l < tcount[k]; l++) {
	    printf (" %7.2f", tsci[k][l]);
	  }
	  printf ("\n");
	}
      }
      /******  Print statements that can be turned on for debugging *****/
      if (DEBUG) {
	/* Set the pixel you want to examine up top */

	if ((i==X1) && (j==Y1)){
	  printf("========== Debug info: iteration %d ====================\n",
		 iter_count);
	  printf("Pixel i,j=%d,%d:    nsamp %d\n",i,j,nsamp);
	  printf("Sci: ");
	  for (k=0; k<nsamp; k++){
	    printf("%  10.6f",sci[k]);
	  }
	  printf("\nDQ: ");
	  for (k=0; k<nsamp; k++){
	    printf("%  5d",dq[k]);
	  }
	  printf("\nTIME: ");
	  for (k=0; k<nsamp; k++){
	    printf("%  10.6f",time[k]);
	  }

	  printf("\n\n Intervals: %d  \n", interval+1);
	  for (idx=0; idx<=interval; idx++){
	    printf("Interval %d, tcount %d\n sci: ",idx+1,tcount[idx]);
	    for (k=0; k<tcount[idx]; k++){
	      printf("%  10.6f",tsci[idx][k]);
	    }
	    printf("\n  dq:");
	    for (k=0; k<tcount[idx]; k++){
	      printf("%  5d",tdq[idx][k]);
	    }
	    printf("\ntime:");
	    for (k=0; k<tcount[idx]; k++){
	      printf("%  10.6f",ttime[idx][k]);
	    }
	    printf("\n...........................\n");
	  }  /* End for loop over idx  */
	}    /* End if DEBUG pixel */
      }      /* End if DEBUG

	     /*****************end debugging***********************/
    

      /*---------------------------------------------------------------
	FIT EACH (non-zero-length) INTERVAL SEPARATELY
	----------------------------------------------------------------*/

      /* We need a unique solution for each interval. */
      for (idx=0;(idx<=interval)&&nrej==0;idx++){
	if (tcount[idx] > 1){
	  /* Compute mean countrate using linear fit and equal weighting 
	     to best find spikes and CRs*/
	  linfit (equal_weight, ttime[idx], tsci[idx], terr[idx], 
		  tdarkandglow[idx], tcount[idx], gain, flat_uncertainty, 
		  &a[idx], &b[idx], &siga[idx], &sigb[idx]);
          if (DEBUG) {
	    printf ("Slope = %8.2f, intercept = %8.2f ", b[idx], a[idx]);
            printf ("(equal weight)\n");
	  }
	      
	  /*---------------------------------------------------------------
	    EXAMINE EACH INTERVAL FOR SPIKES AND CRS
	    ----------------------------------------------------------------*/

	  /* Compute difference of each sample from the local fit */
          naccum = 0;
	  for (k = 0; k <tcount[idx]; k++) {
	    if ((terr[idx][k] != 0) && ((tdq[idx][k] & DQIGNORE) == 0)){
	      diff[k] =(tsci[idx][k]-(a[idx]+b[idx]*ttime[idx][k]))
		/ terr[idx][k];
              naccum +=1;
	    }
	    else{ /*but ignore diffs for already-identified CRs & spikes*/
	      diff[k] = 0;
	    }
	  }
          if (DEBUG) printf ("Accumulated %d diffs\n", naccum);

	  /* Look for and flag new electronic noise spikes and CRs*/
	  nrej=RejSpikes (tsci[idx], terr[idx], tdq[idx], diff, tcount[idx],
			  thresh, &rej);
	  if (nrej > 0) {
	    dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] | BADPIX;
            if (DEBUG) printf ("Rejected %d points\n", nrej);
	  } else if (nrej == 0) {
	    /* Only look for new CR hits if no spike was found */
	    nrej = RejCRs (tdq[idx], diff, tcount[idx], thresh, &rej);
	    if (nrej > 0) {
	      dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] | CR_HIT;
              if (DEBUG) printf ("Rejected %d cosmic rays\n", nrej);
	      number_CRs +=1;
	      if (number_CRs >= max_CRs) {
		found_error = 1;
		dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] - CR_HIT;
		dq[tlookup[idx][rej]] = dq[tlookup[idx][rej]] | HIGH_CURVATURE;
		nrej = 0; /*To break out of the while loop*/
		break;
	      }
	    } else if (nrej == 0) {
	      /* Only check the first read if nothing else has been found */
	      nrej = RejFirstRead(tdq[idx], diff, tcount[idx], thresh);
	      if (nrej > 0){
		dq[tlookup[idx][0]] = dq[tlookup[idx][0]] | BADPIX;
                if (DEBUG) printf ("Rejected first read\n");
	      }
	    }
	  }
	} else if (tcount[idx] == 1) {
	  if ((tdq[idx][0] & CR_HIT) != 0) { /*ignore CR hits if alone*/
            if (ttime[idx][0] == 0) {
              if (DEBUG) printf ("Exposure time is zero\n");
            } else {
	      b[idx]=tsci[idx][0]/ttime[idx][0];
	      sigb[idx]=terr[idx][0]/ttime[idx][0];
            }
	  }
	}   /*endif tcount for this interval*/
      }     /*for each interval*/
    }       /*while new rejections occur*/
  }         /*if no_samples == 1

  /*--------------------------------------------------------------------
           ITERATION COMPLETE: PERFORM FINAL FIT OF EACH INTERVAL
   ---------------------------------------------------------------------*/

  /* Now go back through the intervals and find the best slopes using 
     the optimum weighting */
  int_time = 0;
  sumwts = 0;
  sum = 0;
  if (found_error == 0 && no_samples == 0) {
    for (idx=0; idx<=interval; idx++) {
      if (tcount[idx] > 1) {
	/* Compute mean countrate using linear fit,
	   with optimum weighting this time to best estimate of the slope*/
	linfit (optimum_weight, ttime[idx], tsci[idx], terr[idx], 
		tdarkandglow[idx],tcount[idx], gain, flat_uncertainty, 
		&a[idx], &b[idx], &siga[idx], &sigb[idx]);
        if (nsflag == 1) {
          if (DEBUG) {
	    printf ("Slope = %8.2f, intercept = %8.2f", b[idx], a[idx]);
            printf (" (optimum weight)\n");
          }
        }
	/*also accumulate the integration time and sums  */
	int_time += ttime[idx][tcount[idx]-1] - ttime[idx][0];
        sigsquared = sigb[idx]*sigb[idx];
	if (sigsquared == 0.0) {
          if (DEBUG) printf ("sigb = 0\n");
	} else {
	  sumwts += 1.0/sigsquared;
	  sum    += b[idx]/sigsquared;
        }
      } else if (tcount[idx] == 1) {
	if ((tdq[idx][0] & CR_HIT) != 0) {
          if (ttime[idx][0] == 0.0) {
	    if (DEBUG) printf ("Exposure time is zero\n");
	  } else {
	    b[idx]=tsci[idx][0]/ttime[idx][0];
	    sigb[idx]=terr[idx][0]/ttime[idx][0];
          }
	  k = tlookup[idx][0];
	  int_time += time[k] - time[k-1];
          sigsquared = sigb[idx]*sigb[idx];
          if (sigsquared == 0.0) {
            if (DEBUG) printf ("sigb[idx] = 0 and tcount[idx] = 1\n");
          } else {
	    sumwts += 1/sigsquared;
	    sum    += b[idx]/sigsquared;
          }
	}
      }
    }  /*end for loop over idx*/
  }    /*end test against found_error*/

  /*----------------------------------------------------------------
    COMBINE ALL THE FITS TO GET THE FINAL SOLUTION FOR THE PIXEL
    ---------------------------------------------------------------*/

  /* Set output SCI and ERR values */
  if (no_samples == 0) {
    if ((found_error == 0) && (sumwts > 0.0)) {
      *(out_sci) = sum/sumwts;
      *(out_err) = sqrt(1/sumwts);
      *(out_time) = int_time;
      if (nsflag == 1) {
        if (DEBUG) printf ("Output = %8.2f\n", *(out_sci));
      }
      /* Compute output SAMP  */
      for (idx = 0; idx <= interval; idx++) {
        (*out_samp)+=tcount[idx];
      }
    } else {
      /*Set the flux and errors to the last good read*/
      firstgood = 0;
      lastgood = 0;
      ngood = 0;
      for (k=1; k < nsamp; k++) {
        if ((dq[k] & DQIGNORE) == 0) {
	  firstgood = k;
	  break;
	}
      }
      for (k=firstgood; k < nsamp; k++) {
	if (dq[k] == 0) {
	  lastgood = k;
	  ngood++;
	}
      }
      if (lastgood != firstgood) {
        *(out_sci) = (sci[lastgood] - sci[firstgood]) /
	  (time[lastgood] - time[firstgood]);
	*(out_err) = (err[lastgood] - err[firstgood]) /
	  (time[lastgood] - time[firstgood]);
        *(out_time) = time[lastgood] - time[firstgood];
	*(out_samp) = ngood;
      } else {
	*(out_sci) = 0;
	*(out_err) = 0;
        *(out_time) = 0;
        *(out_samp) = 0;
      }
    }
  } else {
    *(out_err) = 1e5;
  }

  if (DEBUG && i==X1 && j==Y1){
    printf("Output values SCI/ERR/SAMP/TIME %f / %f / %d / %f \n",
	   *out_sci,*out_err,*out_samp,*out_time);
    printf("========== End debug info ====================\n");
  }
  /*--------------------------------------------------------------
                  Clean up and go home
  ----------------------------------------------------------------*/

  free (*tsci);
  free (*terr);
  free (*ttime);
  free (*tdq);
  free (*tdarkandglow);
  free (*tlookup);
  free (diff);
  }



/* LINFIT: Compute a linear fit of the form y = a + bx to a set of
** data points x, y with individual standard deviations sig. Returned
** are a, b and their respective probable uncertainties siga and sigb.
** Taken from Numerical Recipes (with modifications to handle masked data).
**
** Modified to remove mask: only data to be used should be passed in to
** this routine. Optimal weighting (SNR-dependent) may now be used in
** the fit. Calculation of uncertainty now includes contributions
** from Poisson noise (source, dark and glow) and flat uncertainty.
*/

static void linfit (short weight_type, float *x, float *y, float *sig, 
		    float *darkandglow, short ndata, 
		    float gain, float flat_uncert,
		    float *a, float *b, float *siga, float *sigb) {

  /* Local variables */
  int k, i;
  float S, Sx, Sy, Sxx, Sxy, t, wt;
  float snr,power,fit_uncert,dx,dy,ddg,denom;
  float terma, termb, errterms;

  /* Initialize accumulators and results */
  Sx = 0; Sy = 0; Sxx=0; Sxy=0; S=0;
  *a = 0; *b = 0; *siga = 0; *sigb = 0;

   
  /* Check for trivial cases */
  if (ndata == 0){
    (*b)=0.0;
    (*sigb)=1e10;
    return;
  }
  else if (ndata == 1) {
    if (x[0] == 0.0) {
      printf ("x[0] = 0\n");
    } else {
      (*b)    = y[0] / x[0];
      (*sigb) = sig[0] / x[0];
    }
    return;
  }


  if (weight_type == equal_weight){
    power=0;
    snr=0;
  }else{
    /* Determine the correct power term for the optimum weighting.
     * First determine the SNR over the whole interval */
    /* The sig term of the final read is the total rms in the final read
       (Poisson and read) */
    if (sig[ndata-1] > 0.0 ){
      snr=(y[ndata-1]-y[0])/sig[ndata-1];
    }else{
      snr=0.0;
    }
    /* Then use that SNR to select the exponent for the weighting. 
    * These numbers come from a paper by Fixsen (ref. TBA)*/
    if (snr > 100) {
      power =10.0;
    }else{
      if (snr > 50){
	power =6.0;
      }else{
	if(snr > 20){
	  power=3.0;
	}else{
	  if(snr>10){
	    power=1.0;
	  }else{
	    if(snr>5){
	      power=0.4;
	    }else{
	      power=0;
	    }
	  }
	}
      }
    }
  }

  
  /* Accumulate sums */
  for (k = 0; k < ndata; k++) {
    if (sig[k] != 0) {

      wt=fabs(pow(fabs(k-((ndata-1)/2.))/((ndata-1)/2.),power)); 
      /*wt = 1.0 / (sig[k]*sig[k]);*/

      S += wt;
      Sy += y[k] * wt;
      Sx += x[k] * wt;
      Sxx += x[k] * x[k] * wt;
      Sxy += x[k] * y[k] * wt;
    }
  }
  if (S>0){
    dx = x[ndata-1]-x[0];
    dy = y[ndata-1]-y[0];
    ddg = darkandglow[ndata-1]-darkandglow[0];
    denom = (S*Sxx - (Sx*Sx));

    if (denom < 1e-2){
      /*Avoid dividing by zero or other problematic numerical issues*/
      denom = 1e-2;
    }

    if (ndata == 2){ 
      *b = dy/dx;
      *a = y[0] - (*b)*x[0];
      *siga = 1.0; /*not strictly correct, but not really used*/
      fit_uncert = 0;
      *sigb = sqrt((sig[0]/x[0])*(sig[0]/x[0]) + (sig[1]/x[1])*(sig[1]/x[1]));

    }else{ /* we can do a proper calculation */
      *b = (S*Sxy - Sx*Sy)/denom;
      *a = (Sxx*Sy - Sx*Sxy)/denom;
      *siga = sqrt(Sxx/denom);
      fit_uncert = sqrt(S/denom);
      *sigb = fit_uncert;
    }

    /* Contributions to the slope uncertainty are:
         - the fit uncertainty = fit_uncert
         - the Poisson noise from dark and glow = ddg

	 However, it's only worth doing this calculation if we are not down
    in the read noise */

    terma = (gain*fit_uncert*dx)*(gain*fit_uncert*dx);
    termb =  gain*ddg;

    errterms = terma + termb;


    if (errterms > 0){
      *sigb = (sqrt(terma + termb)/dx)/gain;
    }

  }else{ /*No points had any weight*/
    *a=0.0;
    *b=0.0;
    *sigb=1e9;
    *siga=1e9;
  }
}



/* REJSPIKES: Find and flag electronic noise spikes. These show up as
** large positive or negative deviations for a single sample relative to
** the samples on either side of it.
** Rewritten by MWR 4/2007
*/

static int RejSpikes (float *tsci, float*terr, short *dq, float *diff,
		       short nsamp, float thresh, int *max_samp) {

  /* Local variables */
  int   i,k, nrej;		/* loop index */
  float max_diff; 	/* max sample deviation */
  
  max_diff = 0.0; (*max_samp) = 0;
  nrej = 0;
  
  /* Loop over samples skipping the first and last samples*/
  /* The first sample is treated separately and a spike in the last sample */
  /* is treated as a cosmic ray. */
  for (k = 2; k < nsamp; k++) {
    if (((dq[k-1] & (DQIGNORE | CR_HIT | BADPIX)) == 0) &&
        ((dq[k] & (DQIGNORE | CR_HIT | BADPIX)) == 0) &&
        ((dq[k+1] & (DQIGNORE | CR_HIT | BADPIX)) == 0)) {
      
      /* Check for positive or negative spikes */
      /* Here we look for a large difference that has small differences */
      /* on both sides. This spike is just ignored in the fit and */
      /* doesn't affect further reads. */
      if (diff[k-1] - diff[k-2] > thresh &&
	  diff[k-1] - diff[k]   > thresh &&
	  diff[k-2] < 0 &&
	  diff[k]   < 0){
	if (fabs(diff[k-1]) > max_diff) {
	  max_diff = diff[k-1];
	  (*max_samp) = k-1;
	  nrej++;
	}
      }
      if (diff[k-2] - diff[k-1] > thresh &&
	  diff[k] - diff[k-1]   > thresh &&
	  diff[k-2] > 0 &&
	  diff[k]   > 0){
	if (fabs(diff[k-1]) > max_diff) {
	  max_diff = diff[k-1];
	  (*max_samp) = k-1;
	  nrej++;
	}
      }
    }
  }
  /* If a spike was found, tell the caller*/
  if (nrej > 0){
    return 1;
  }else{
    return 0;
  }
}

/* REJFIRSTREAD: Check for spikes in the first read of the interval only.
** This was formerly part of RejSpikes. 
*/
static int RejFirstRead(short *dq, float *diff, short nsamp, float thresh) {
  /* Local variables */
  int   nrej;		/* return value */
  
  nrej = 0;
  

  if (((dq[0] & (DQIGNORE | CR_HIT | BADPIX)) == 0) &&
      ((dq[1] & (DQIGNORE | CR_HIT | BADPIX)) == 0)){
/* first read of interval is not flagged already */
    if (fabs(diff[0]-diff[1]) > thresh){ /* a positive or negative spike */
	nrej++;
    }
  }
  /* If a spike was found, flag the sample that has the max deviation */
  if (nrej > 0){
    return 1;
  }else{
    return 0;
  }
}

/* REJCRS: Find and flag a Cosmic Ray hits. This is done by looking for
** large negative-to-positive going changes in the sample values from one
** sample to the next.
** Rewritten by MWR 4/2007
** Further rewritten to return 1 or 0 depending on whether a CR was
** found, and return the associated cr_index in the call. VGL, Jun 2007
*/

static int RejCRs (short *dq, float *diff, short nsamp, float thresh,
		    int *cr_index) {
  
  /* Local variables */
  int k,i;			/* loop index */
  float prev_diff;	/* last useable difference value */
  float max_diff; /*the largest jump */
  prev_diff = 0;
  max_diff = 0;

  /* Loop over samples */
  for (k = 1; k < nsamp; k++) {
    if ((dq[k-1] & (DQIGNORE | CR_HIT | BADPIX)) == 0)
      prev_diff = diff[k-1];
    if ((dq[k] & (DQIGNORE | CR_HIT | BADPIX)) == 0) {
      /* A large positive increase in the difference is used as */
      /* a sign of a CR. */
      if (diff[k]-prev_diff > thresh) {
	if (diff[k]-prev_diff > max_diff){
	  max_diff = diff[k]-prev_diff;
	  (*cr_index) = k;
	}
	prev_diff = diff[k];
      }
    }
  }
  if (max_diff != 0){
    return 1;
  }else{
    return 0;
  }
}

/* This subroutine determines the linear dark current and amp glow components
   for each read of a pixel. This is called when temperature dependent
   darks are used and we have access to the seperate linear dark and amp
   glow components. */
static void GetDarkAndGlow(short operation, char dark_ref_name[SZ_NAME+1],
                           NicInfo *nic, short first_read, short last_read,
                           float *time, short i, short j,
                           float *Poisson_counts)
{
  void n_amulk_nooverwrite (SingleNicmosGroup *,SingleNicmosGroup *, float);
  void n_aadd_nooverwrite (SingleNicmosGroup *,SingleNicmosGroup *,
                           SingleNicmosGroup *);

  static	SingleNicmosGroup lin_dark;
  static	SingleNicmosGroup amp_glow;
  float  ADUsFromDark;
  float  ADUsFromGlow;
  float  TotalADUs;
  static		SingleNicmosGroup *lin_dark_ptr;
  static		SingleNicmosGroup *amp_glow_ptr;
  FloatHdrData xdark;
  int k;
  float lin_scale;
  float amp_scale;
  
  if (operation == clean_up){ 
    lin_dark_ptr = NULL;
    amp_glow_ptr = NULL;    
    freeSingleNicmosGroup (&lin_dark);
    freeSingleNicmosGroup (&amp_glow);
  }else{
    if (lin_dark_ptr == NULL){
      if (DEBUG){
	printf("init dark current\n");
      }
      /* Open the dark reference file  */
      initFloatHdrData (&xdark);
      
      if (getFloatHD (dark_ref_name, "LIN", 1, &xdark)) {
	sprintf (MsgText, "cridcalc Error getting LIN group of dark image");
	n_error (MsgText);
	return;
      }
      
      /*  Need to convert to a SingleNicmosGroup  */
      if (n_copytoSingleNicmosGroup (&lin_dark, xdark, nic, 1)) {
	sprintf (MsgText, "cridcalc Error converting LIN extension to SNG");
	n_error (MsgText);
	  return;
      }
      lin_dark_ptr = &lin_dark;
      freeFloatHdrData (&xdark);
	
      /* Get the AMP extension from reference file  */
      initFloatHdrData (&xdark);
	
      if (getFloatHD (dark_ref_name, "AMPGLOW", 1, &xdark)) {
	sprintf (MsgText, "cridcalc Error getting AMPGLOW group of DARK" \
                           " image");
	n_error (MsgText);
	return;
      }
      
      /* Convert to SNG  */
      if (n_copytoSingleNicmosGroup (&amp_glow, xdark, nic, 1)) {
	sprintf (MsgText,"cridcalc Error converting AMPGLOW extension to SNG");
	n_error (MsgText);
	return;
      }
      amp_glow_ptr=&amp_glow;
      freeFloatHdrData (&xdark);

    }   
    
    for (k=first_read; k<= last_read; k++){

      /* Multiply by the delta time  */
        
    /*this takes into account temperature dependent amp and lin dark files*/    
    if(nic->tfbtemp != -1 && (nic->ampScale != 1 || nic->linScale !=1) ){
       	lin_scale=nic->linScale;
        amp_scale=nic->ampScale;
    }  else {
        lin_scale=1.;
        amp_scale=1.;
    }
    
     ADUsFromDark = Pix(lin_dark.sci.data,i,j) * time[k] * lin_scale;
      
      /* Multiply AMPGLOW image by the number of reads  */
      /* There are k+1 reads because the 0th read has one read */
      /* So when K==1 the number of reads is 2.         */
      	ADUsFromGlow = Pix(amp_glow.sci.data,i,j) * (k+1) * amp_scale;
      
      /* Now add the ampglow and linear dark components together */
      
      Poisson_counts[k] = ADUsFromDark + ADUsFromGlow;

    }
  }
}


/* This subroutine is used to estimate the dark and glow component of the
   dark when we only have a super dark. It finds the set of reads that
   all have the same delta time to avoid shading as much as possible. It then
   deteremines the average rate of each pixel and stores that in the rate
   structure. When called it will return the tot_ADUs for each read of a
   given pixel. */
static void EstimateDarkandGlow(short operation, NicInfo *nic, short nsamp,
                                short row, short col, float *time,
                                float *tot_ADUs)
{
  SingleNicmosGroup lastwithsamedelta;
  SingleNicmosGroup firstwithsamedelta; 
  static SingleNicmosGroup rate;
  static SingleNicmosGroup *rate_ptr;
  float pre_exp_time,prev_delta_time,prev_exp_time,delta_time;
  int   i,j,start_even_intervals,end_even_intervals;

  if (operation == clean_up){ 
    rate_ptr = NULL;
    freeSingleNicmosGroup (&rate);
  }else{
    if (rate_ptr == NULL){
      prev_exp_time = nic->dtimes[0];
      prev_delta_time =0.0;
      start_even_intervals = -100;
      end_even_intervals = -100;
      for (i=1; i<=nsamp-1; i++){
	if (nic->dtimes[i] > prev_exp_time){ /* increasing exp times */
	  delta_time = nic->dtimes[i] - nic->dtimes[i-1];
	  if (abs(delta_time-prev_delta_time) < 0.01 &&
              (start_even_intervals < 0.0)){
	    start_even_intervals = i-1;
	    end_even_intervals = nsamp-1;
	  }
	}else{/* decreasing exp times */
	  delta_time = nic->dtimes[i-1] - nic->dtimes[i];
	  if (i != 1 && abs(delta_time-prev_delta_time)>0.01 &&
                        (end_even_intervals < 0.0)){
	    start_even_intervals = 0;
	    end_even_intervals =i-1;
	  }
	}
	prev_delta_time= delta_time;
	prev_exp_time = nic->dtimes[i];
      }

      /* Make sure we found some evenly spaced intervals */
      if (start_even_intervals == -100){
	if (nic->dtimes[nsamp-1] > nic->dtimes[0]){
	  /*then the exposure times are increasing*/
	  start_even_intervals=nsamp-2;
	  end_even_intervals=nsamp-1;
	}else{ /* they are decreasing*/
	  start_even_intervals=2;
	  end_even_intervals=1;
	}
      }

      initSingleNicmosGroup (&firstwithsamedelta);
      initSingleNicmosGroup (&lastwithsamedelta);
      initSingleNicmosGroup (&rate);
      rate_ptr = &rate;
      n_getStaticDarkImage (nic, &firstwithsamedelta, start_even_intervals); 
      n_getStaticDarkImage (nic, &lastwithsamedelta, end_even_intervals); 
      n_getStaticDarkImage (nic, &rate, end_even_intervals); 
      for (i=0;i<firstwithsamedelta.sci.data.nx; i++){
	for (j=0;j<firstwithsamedelta.sci.data.ny; j++){
	  Pix(rate.sci.data,i,j) = (Pix(lastwithsamedelta.sci.data,i,j) -
                                    Pix(firstwithsamedelta.sci.data,i,j))/
	    (nic->dtimes[end_even_intervals] -
             nic->dtimes[start_even_intervals]);

	}
      }
    }
    /* we have created the rate array */
    for (i=0; i <nsamp; i++){
      tot_ADUs[i] = time[i] * Pix(rate.sci.data,row,col);
    }
  }
}
