# include <stdio.h>
# include <stdlib.h>	/* malloc */
# include <string.h>    /* strchr */
# include <ctype.h>

# include <xsynphot.h>		/* for c_phopar */

# include <hstio.h>
# include "../wf3.h"
# include "../wf3info.h"
# include "../wf3err.h"

# define NPHOT      4		/* size of phot returned by c_phopar */
# define ARR_SIZE   10000	/* size of arrays used for throughputs */

static void  Phot2Obs (char *, char *);

/* This routine calls synphot functions to compute values associated with
   the absolute flux calibration of the science data and populates those
   values into a set of photometry keywords in the science image header.

   Warren Hack, 1998 June 12:
   	Initial ACS version.  
   Howard Bushouse, 2000 Aug 29:
	Initial WFC3 version.
   H.Bushouse, 2002 June 17:
	Removed gain normalization of PHOTFLAM keyword value, to be consistent
	with image datatype after flat-field calibration. Also, increased
	ARR_SIZE to match the default value used by SYNPHOT (in accordance
	with CALACS changes).
   H.Bushouse, 2008 Nov 5:
	Added computation of PHOTFNU keyword value, to be consistent with the
	WFC3 IR photcorr process. Also removed some old ACS-specific code that
	is not used for WFC3.
*/

int doPhot (WF3Info *wf32d, SingleGroup *x) {

/* arguments:
WF3Info *wf32d     i: calibration switches, etc
SingleGroup *x    io: image to be calibrated; primary header is modified
*/

	extern int status;

	PhotInfo phot;		/* QE and filter throughput, vs wavelength */
	float photfnu;
	
	float photkey[NPHOT];	/* values of photometry keywords */
	char  photmode[SZ_LINE+1], obsmode[SZ_LINE+1];
	char  grafpath[1025];

	int GetKeyStr (Hdr *, char *, int, char *, char *, int);
	int PutKeyFlt (Hdr *, char *, float, char *);

	if (wf32d->photcorr == DUMMY)
	    return (status);

	/* Extract photmode from sci extension header */
	if (GetKeyStr (&x->sci.hdr, "PHOTMODE",USE_DEFAULT,"",photmode,SZ_LINE))
	    return (status);

	/* Add commas to PHOTMODE string and change all strings to
	** lower case for use in synphot. */
	Phot2Obs (photmode, obsmode);
	if (wf32d->verbose) {
	    sprintf (MsgText, "Created SYNPHOT obsmode of: %s", obsmode);
	    trlmessage (MsgText);
	}

	/* Initialize photometry keyword structures */
	phot.nelem  = ARR_SIZE;
	phot.f_wl   = calloc (phot.nelem, sizeof(float));
	phot.f_thru = calloc (phot.nelem, sizeof(float));
	phot.f_err  = calloc (phot.nelem, sizeof(float));

	if (phot.f_wl == NULL || phot.f_thru == NULL || phot.f_err == NULL)
	    return (status = OUT_OF_MEMORY);

	/* Pass obsmode string to synphot to generate throughput array */
	c_getbandx (obsmode, wf32d->graph.name, wf32d->comp.name, True,
		    phot.nelem, phot.f_wl, phot.f_thru, phot.f_err);
	if (c_iraferr()) {
	    trlerror ("Error return from c_getbandx");
	    return (status = ERROR_RETURN);
	}

	/* Compute photometry values. */
	c_phopar (phot.nelem, phot.f_wl, phot.f_thru, photkey);
	if (c_iraferr()) {
	    trlerror ("Error return from c_phopar");
	    return (status = ERROR_RETURN);
	}

	c_listpath (obsmode,wf32d->graph.name,wf32d->comp.name,grafpath,1025);
	trlmessage (grafpath);

	/* Add this information as a HISTORY comment */
	addHistoryKw (&x->sci.hdr, grafpath);
	if (hstio_err())
	    return (status = HEADER_PROBLEM);

	if (wf32d->verbose) {
	    sprintf (MsgText, "Computed PHOTFLAM value of %g", photkey[0]);
	    trlmessage (MsgText);
	}

	/* Update the photometry keyword values in the SCI header. */
	if (PutKeyFlt (&x->sci.hdr, "PHOTFLAM", photkey[0],
		       "inverse sensitivity"))
	    return (status);

	if (PutKeyFlt (&x->sci.hdr, "PHOTZPT", photkey[1], "zero point"))
	    return (status);

	if (PutKeyFlt (&x->sci.hdr, "PHOTPLAM", photkey[2], "pivot wavelength"))
	    return (status);

	if (PutKeyFlt (&x->sci.hdr, "PHOTBW", photkey[3], "RMS bandwidth"))
	    return (status);

	photfnu = 3.33564e+4 * photkey[0] * photkey[2]*photkey[2];

	if (PutKeyFlt (&x->sci.hdr, "PHOTFNU", photfnu, "inverse sensitivity"))
	    return (status);

	free (phot.f_wl);
	free (phot.f_thru);
	free (phot.f_err);

	return (status);
}

/* This function converts the PHOTMODE string into an OBSMODE string
   suitable for use with synphot functions.
	PHOTMODE - all upper case component names separated by blanks
	OBSMODE - all lower case names separated by commas
   NOTE: The operation occurs **in-place** on the photmode string.
*/
static void Phot2Obs (char *photmode, char *obsmode) {

	char blank = 32, comma = 44;
	int i, len, c1;

	len = strlen (photmode);

	for (i=0; i < len; i++) {
	     c1 = photmode[i];
	     if (c1 == blank) {
		 obsmode[i] = comma;
	     } else
		 obsmode[i] = tolower(c1);
	}
	obsmode[len] = '\0';
}

