# include <stdlib.h>
# include <string.h>
# include <ctype.h>

# include <xsynphot.h>

# include <hstio.h>	/* defines HST I/O functions */
# include "../wf3.h"
# include "../wf3info.h"
# include "../wf3err.h"

extern int status;

/* PHOTCALC: Store the photometry parameter values in the global header
** keywords PHOTMODE, PHOTFLAM, PHOTFNU, PHOTZPT, PHOTPLAM, and PHOTBW.
** 
** The input data are NOT modified (only the global header is modified).
**
** Revision history:
** H.Bushouse	Oct. 2000	Initial CALNICA to CALWF3 port.
** H.Bushouse	19-Nov-2001	Modified to track CALACS changes: Call
**				Synphot library functions to compute
**				photometry parameters, instead of reading
**				from the PHOTTAB ref table. Copied from
**				CALWF3 UVIS routine.
*/

# define NPHOT		4	/* size of phot returned by c_phopar */
# define ARR_SIZE	10000	/* size of arrays used for throughputs */

static void Phot2Obs (char *, char *);

int photcalc (WF3Info *wf3, MultiNicmosGroup *input) {

/* Arguments:
**	wf3	 i: WFC3 info structure
**	input	io: input image
*/

	/* Local variables */
	PhotInfo phot;
	float photfnu;

	float photkey[NPHOT];	/* values of photometry keywords */
	char  photmode[SZ_LINE+1], obsmode[SZ_LINE+1];
	char  grafpath[1025];

	/* Function definitions */
	int GetKeyStr (Hdr *, char *, int, char *, char *, int);
	int PutKeyFlt (Hdr *, char *, float, char *);
	void PrSwitch (char *, int);

	if (wf3->photcorr == PERFORM) {

	/* Extract photmode from sci extension header */
	if (GetKeyStr (input->group[0].globalhdr, "PHOTMODE", USE_DEFAULT, "",
		       photmode, SZ_LINE))
	    return (status);

	/* Convert PHOTMODE string into synphot OBSMODE syntax */
	Phot2Obs (photmode, obsmode);
	if (wf3->verbose) {
	    sprintf (MsgText, "Created SYNPHOT obsmode of: %s", obsmode);
	    trlmessage (MsgText);
	}

	/* Initialize photometry keyword structures */
	phot.nelem  = ARR_SIZE;
	phot.f_wl   = calloc (phot.nelem, sizeof(float));
	phot.f_thru = calloc (phot.nelem, sizeof(float));
	phot.f_err  = calloc (phot.nelem, sizeof(float));

	if (phot.f_wl == NULL || phot.f_thru == NULL || phot.f_err == NULL)
	    return (status = OUT_OF_MEMORY);

	/* Pass obsmode string to synphot to generate throughput array */
	c_getbandx (obsmode, wf3->graph.name, wf3->comp.name, True,
		    phot.nelem, phot.f_wl, phot.f_thru, phot.f_err);
	if (c_iraferr()) {
	    trlerror ("Error return from c_getbandx");
	    return (status = ERROR_RETURN);
	}

	/* Compute photometry values */
	c_phopar (phot.nelem, phot.f_wl, phot.f_thru, photkey);
	if (c_iraferr()) {
	    trlerror ("Error return from c_phopar");
	    return (status = ERROR_RETURN);
	}

	c_listpath (obsmode, wf3->graph.name, wf3->comp.name, grafpath, 1025);
	trlmessage (grafpath);

	/* Add this information as a HISTORY comment */
	addHistoryKw (input->group[0].globalhdr, grafpath);
	if (hstio_err())
	    return (status = HEADER_PROBLEM);

	/* Update the photometry keyword values in the header */
	if (PutKeyFlt (input->group[0].globalhdr, "PHOTFLAM", photkey[0], ""))
	    return (status);

	if (PutKeyFlt (input->group[0].globalhdr, "PHOTZPT", photkey[1], ""))
	    return (status);

	if (PutKeyFlt (input->group[0].globalhdr, "PHOTPLAM", photkey[2], ""))
	    return (status);

	if (PutKeyFlt (input->group[0].globalhdr, "PHOTBW", photkey[3], ""))
	    return (status);

	photfnu = 3.33564e+4 * photkey[0] * photkey[2]*photkey[2];

	if (PutKeyFlt (input->group[0].globalhdr, "PHOTFNU", photfnu, ""))
	    return (status);

	free (phot.f_wl);
	free (phot.f_thru);
	free (phot.f_err);

	PrSwitch ("photcorr", COMPLETE);
	}

	/* Successful return */
	return (status = 0);
}

/* This function converts the PHOTMODE string into an OBSMODE string
   suitable for use with synphot functions.
        PHOTMODE - all upper case component names separated by blanks
        OBSMODE - all lower case names separated by commas
   NOTE: The operation occurs **in-place** on the photmode string.
*/
static void Phot2Obs (char *photmode, char *obsmode) {

        char blank = 32, comma = 44;
        int i, len, c1;

        len = strlen (photmode);

        for (i=0; i < len; i++) {
             c1 = photmode[i];
             if (c1 == blank) {
                 obsmode[i] = comma;
             } else
                 obsmode[i] = tolower(c1);
        }
        obsmode[len] = '\0';
}

