#!/usr/bin/env python

""" runwf3driz.py - Module to control operation of MultiDrizzle to 
        remove distortion and combine WFC3 images in the pipeline.

USAGE: runwf3driz.py inputFilename 

Alternative USAGE: python
                   from runwf3driz import *
                   run(inputFilename)
IRAF Usage:
    epar runwf3driz
    runwf3driz input mode=h
    
*** INITIAL VERSION

W.J. Hack  9 Oct 2009

H.A. Bushouse 2 Dec 2009: Comment out IF block that reads DRIZCORR 
setting from drz file (if it already exists), so that DRIZCORR is 
always read from the first member of the asn table.  

H.A. Bushouse, 27 Jan 2010: Fixed to make logic work with singletons,
which involved modifications to the use of fname and mname so that
mname is always populated for both singletons and associations.
Trac ticket #506.

"""

# Import standard Python modules
from __future__ import division # confidence high
import os, sys, string,time, shutil
import glob

# Import necessary IRAF modules for CALACS
import pyraf
import iraf

yes = iraf.yes
no = iraf.no

# For testing purposes only...
#if os.environ.has_key('stsdas'):
#    iraf.set(stsdas = os.environ['stsdas'])

from pyraf.iraf import stsdas,hst_calib,dither

import multidrizzle
from stsci.tools import fileutil, asnutil

# Import local modules
import pyfits

# Local variables 
__version__ = "1.2.0"
__vdate__ = "(27-Jan-2010)"
__bitvalue__ = 8192 + 2 + 128 +256        # default value for MultiDrizzle usage
#default marker for trailer files
__trlmarker__ = '*** WF3 MultiDrizzle Processing Version '+__version__+__vdate__+'***\n'

# History:
# Version 1.0.0 - Derived from v2.1.6 of acs.runcaldriz to only run MultiDrizzle
# Version 1.1.0 - Always read DRIZCORR from first ASN member, not from drz file
# Version 1.2.0 - Fixed logic to read DRIZCORR from singletons
    
# Function to control execution from the shell.
def _main():

    _nargs = len(sys.argv)
    # Check the command line arguments
    if _nargs < 1:
        print "syntax: runwf3driz.py inputFilename"
        sys.exit()
    
    # Invoke the module function
    try:
        run(sys.argv[1])
    except Exception, errorobj:
        print str(errorobj)
        print "ERROR: Cannot run MultiDrizzle on %s." % sys.argv[1]
        raise Exception, str(errorobj)
        
    sys.exit()


def run(inFile):
    """ Run MultiDrizzle on input file/ASN table 
        using default values for MultiDrizzle parameters.
    """
    # Open the input file
    try:
        # Make sure given filename is complete and exists...
        inFilename = fileutil.buildRootname(inFile,ext=['.fits'])
        if not os.path.exists(inFilename):
            print "ERROR: Input file - %s - does not exist." % inFilename
            return
    except TypeError:
        print "ERROR: Inappropriate input file."
        return
        

    # Initialize for later use...
    _mname = None
    # Check input file to see if [DRIZ/DITH]CORR is set to PERFORM
    _indx_asn = inFilename.find('_asn')
    if _indx_asn > 0:
        # We are working with an ASN table.
        # Use PyDrizzle's code to extract filename
        inFilename = _lowerAsn(inFilename)
        _asndict = asnutil.readASNTable(inFilename,None,prodonly=False)
        _fname = fileutil.buildRootname(string.lower(_asndict['output']),ext=['_drz.fits'])
        _cal_prodname = string.lower(_asndict['output'])
        
        # Retrieve the first member's rootname for possible use later
        _fimg = pyfits.open(inFilename)
        for name in _fimg[1].data.field('MEMNAME'):
            if name[-1] != '*':
                _mname = string.lower(string.split(name,'\0',1)[0])
                break
        _fimg.close()
        del _fimg
    
    else:
        # Check to see if input is a _RAW file
        # If it is, strip off the _raw.fits extension...
        _indx = string.find(inFilename,'_raw')
        if _indx < 0: _indx = len(inFilename)
        # ... and build the CALWF3 product rootname.
        _mname = fileutil.buildRootname(inFilename[:_indx])
        _cal_prodname = inFilename[:_indx]
        # Reset inFilename to correspond to appropriate input for
        # drizzle: calibrated product name.
        inFilename = _mname
        
        if _mname == None:
            errorMsg = 'Could not find calibrated product!' 
            raise Exception,errorMsg

    # Create trailer filenames based on ASN output filename or
    # on input name for single exposures
    _indx_raw = inFile.find('_raw')
    
    if _indx_raw > 0:
        # Output trailer file to RAW file's trailer
        _trlroot = inFile[:_indx_raw]
    elif _indx_asn > 0:
        # Output trailer file to ASN file's trailer, not product's trailer
        _trlroot = inFile[:_indx_asn]
    else:
        # Default: trim off last suffix of input filename
        # and replacing with .tra
        _indx = inFile.rfind('_')
        if _indx > 0:
            _trlroot = inFile[:_indx]
        else:
            _trlroot = inFile

    _trlfile = _trlroot + '.tra'

    # Open product and read keyword value
    # Check to see if product already exists...
    dkey = 'DRIZCORR'  
    """
    if fileutil.findFile(_fname):
        # If product exists, try to read in value 
        # of DITHCORR keyword to guide processing...
        _fimg = pyfits.open(_fname)
        _phdr = _fimg['PRIMARY'].header

        if _phdr.has_key(dkey) > 0:
            dcorr = _phdr[dkey]
        else:
            dcorr = None

        # Done with image
        _fimg.close()
        del _fimg    
    else:
    """
    # ...if product does NOT exist, interrogate input file
    # to find out whether 'dcorr' has been set to PERFORM
    if _mname :
        _fimg = pyfits.open(fileutil.buildRootname(_mname,ext=['_raw.fits']))
        _phdr = _fimg['PRIMARY'].header
        if _phdr.has_key(dkey) > 0:
            dcorr = _phdr[dkey]
        else:
            dcorr = None
        _fimg.close()
        del _fimg
    else:
        dcorr = None
     
    time_str = _getTime()
    _tmptrl = _trlroot + '_tmp.tra'
    _drizfile = _trlroot + '_pydriz.tra'
    
    _new_asn = None
    if dcorr == 'PERFORM':    
        if inFilename.find('_asn.fits') == -1:
            # Working with a singleton
            # However, we always want to make sure we always use
            # a calibrated product as input, if available.
            _infile = fileutil.buildRootname(_cal_prodname)
            _cal_prodname = _infile
        else:
            # Working with an ASN table...
            # Need to insure that product name is lower case
            _infile = inFilename
            # We want to keep the original specification of the calibration 
            # product name, though, not a lower-case version...
            _cal_prodname = inFilename
            _new_asn = _infile
        
        # Run MultiDrizzle and send its processing statements to _trlfile
        _pyver = multidrizzle.__version__ 
        # Create trailer marker message for start of MultiDrizzle processing
        _trlmsg = _timestamp('MultiDrizzle started ')
        _trlmsg = _trlmsg+ __trlmarker__
        _trlmsg = _trlmsg + '%s: Processing %s with MultiDrizzle Version %s\n' % (time_str,_infile,_pyver)   
        print _trlmsg

        # Write out trailer comments to trailer file...
        ftmp = open(_tmptrl,'w')
        ftmp.writelines(_trlmsg)
        ftmp.close()
        _appendTrlFile(_trlfile,_tmptrl)
        
        _pyd_err = _trlroot+'_pydriz.stderr'

        try:
            # Run mdrizzle through the IRAF interface now...
            iraf.unlearn('multidrizzle')
            iraf.multidrizzle.clean = yes
            iraf.multidrizzle.mdriztab = yes
            iraf.multidrizzle(input=_infile,Stdout=_drizfile,Stderr=_pyd_err)
            os.remove(_pyd_err)
        except Exception, errorobj:
            _appendTrlFile(_trlfile,_drizfile)
            _appendTrlFile(_trlfile,_pyd_err)
            _ftrl = open(_trlfile,'a')
            _ftrl.write('ERROR: Could not complete MultiDrizzle processing of %s.\n' % _infile) 
            _ftrl.write(str(sys.exc_type)+': ')
            _ftrl.writelines(str(errorobj))
            _ftrl.write('\n')
            _ftrl.close()
            print 'ERROR: Could not complete MultiDrizzle processing of %s.' % _infile
            raise Exception, str(errorobj)
        
        # Now, append comments created by PyDrizzle to CALACS trailer file
        print 'Updating trailer file %s with MultiDrizzle comments.' % _trlfile
        _appendTrlFile(_trlfile,_drizfile)
        
        # Save this for when PyFITS can modify a file 'in-place'
        # Update calibration switch
        _fimg = pyfits.open(_cal_prodname,mode='update')
        _fimg['PRIMARY'].header.update(dkey,'COMPLETE')
        _fimg.close()
        del _fimg

        # Enforce pipeline convention of all lower-case product
        # names
        _prodlist = glob.glob('*drz.fits')
        for _prodname in _prodlist:
            _plower = _prodname.lower()
            if _prodname != _plower:  os.rename(_prodname,_plower)
                
    else:
        # Create default trailer file messages when MultiDrizzle is not
        # run on a file.  This will typically apply only to BIAS,DARK
        # and other reference images.
        # Start by building up the message...
        _trlmsg = _timestamp('MultiDrizzle skipped ')
        _trlmsg = _trlmsg + __trlmarker__
        _trlmsg = _trlmsg + '%s: MultiDrizzle processing not requested for %s.\n' % (time_str,inFilename)
        _trlmsg = _trlmsg + '       MultiDrizzle will not be run at this time.\n'
        print _trlmsg
        
        # Write message out to temp file and append it to full trailer file
        ftmp = open(_tmptrl,'w')
        ftmp.writelines(_trlmsg)
        ftmp.close()
        _appendTrlFile(_trlfile,_tmptrl)
    
    _fmsg = None    
    # Append final timestamp to trailer file...
    _final_msg = '%s: Finished processing %s \n' % (time_str,inFilename)   
    _final_msg += _timestamp('MultiDrizzle completed ')
    _trlmsg += _final_msg
    ftmp = open(_tmptrl,'w')
    ftmp.writelines(_trlmsg)
    ftmp.close()
    _appendTrlFile(_trlfile,_tmptrl)
    
    # If we created a new ASN table, we need to remove it
    if _new_asn != None:
        os.remove(_new_asn)
    
    # Provide feedback to user
    print _final_msg

def _lowerAsn(asnfile):
    """ Create a copy of the original asn file and change
        the case of all members to lower-case.
    """
    # Start by creating a new name for the ASN table
    _indx = asnfile.find('_asn.fits')
    _new_asn = asnfile[:_indx]+'_pipeline'+asnfile[_indx:]
    if os.path.exists(_new_asn) == yes:
        os.remove(_new_asn)
    # copy original ASN table to new table 
    shutil.copy(asnfile,_new_asn)
    
    # Open up the new copy and convert all MEMNAME's to lower-case
    fasn = pyfits.open(_new_asn,'update')
    for i in xrange(len(fasn[1].data)):
        fasn[1].data[i].setfield('MEMNAME',fasn[1].data[i].field('MEMNAME').lower())
    fasn.close()
    
    return _new_asn
    
def _appendTrlFile(trlfile,drizfile):
    """ Append drizfile to already existing trlfile from CALACS.
    """
    # Open already existing CALWF3 trailer file for appending
    ftrl = open(trlfile,'a')
    # Open MultiDrizzle trailer file
    fdriz = open(drizfile)
    
    # Read in drizzle comments
    _dlines = fdriz.readlines()
    
    # Append them to CALWF3 trailer file
    ftrl.writelines(_dlines)
    
    # Close all files
    ftrl.close()
    fdriz.close()
    
    # Now, clean up MultiDrizzle trailer file
    os.remove(drizfile)


def _timestamp(_process_name):
    """Create formatted time string recognizable by OPUS."""
    _prefix= time.strftime("%Y%j%H%M%S-I-----",time.localtime())
    _lenstr = 60 - len(_process_name)
    return _prefix+_process_name+(_lenstr*'-')+'\n'
    
def _getTime():
    # Format time values for keywords IRAF-TLM, and DATE 
    _ltime = time.localtime(time.time())
    time_str = time.strftime('%H:%M:%S (%d-%b-%Y)',_ltime)
    
    return time_str

if __name__ == "__main__":
    _main() 
