# flall -- do all steps to make a UNI file
# This script takes as input a list of flat fields and produces one
# image in RSDP UNI file format.  Optional inputs are an external flat
# field (for filling in a corner of an f/48 internal flat) and a mask
# table (for interpolating along vertical strips over fingers, etc.)
# The images and the focutility packages must be loaded before running
# this task.
# All input images are assumed to be 512 x 1024.  If there is more than
# one input image, they will be added together using IMCOMBINE.  If an
# external flat (INPUT2) was specified, FLCORNER will be run to splice in
# a corner of that image.  A portion of the border will be trimmed off
# using FLBORDER.  If a mask table (INTABLE) was specified, FLVMASK will
# be run to fill in the masked regions.  MEDIAN and GAUSS can be run to
# smooth the image.  BLKREP is then used to "dezoom".  The reciprocal is
# taken using RSDPUNIX, which also converts to proper UNI file format.
# Finally, the result is normalized to unit average within a central region.
#
# Phil Hodge, 13-Mar-1990  Task created.

procedure flall (input, input2, intable, output)

string	input = "" {prompt = "input images"}
string	input2 = "" {prompt = "name of f/48 external flat"}
string	intable = "" {prompt = "name of mask table"}
string	output = "" {prompt = "name of output UNI file"}

real	y1 = 456.4 {prompt = "y intercept of lower line", min=0.}
real	y2 = 656.4 {prompt = "y intercept of upper line", min=2.}
real	slope = 2.446 {prompt = "slope of lines", min=0.1}
int	left_b = 5 {prompt = "size of left border", min=0}
int	right_b = 18 {prompt = "size of right border", min=0}
int	bottom_b = 6 {prompt = "size of bottom border", min=0}
int	top_b = 6 {prompt = "size of top border", min=0}
bool	medfilt = yes {prompt = "perform median filtering?"}
int	width = 7 {prompt = "width in X for median filter", min=1}
string	sm_option = "gauss" {prompt = "type of smoothing", \
			enum = "gauss|boxcar|none"}
real	sigma = 7.6439 {prompt = "sigma for Gaussian filter", min=0.}
real	nsigma = 2.75 {prompt = "width of Gaussian filter, unit=sigma", \
			min = 0.}
int	box = 9 {prompt = "box size for boxcar smoothing", min=1}
real	threshold = 1.e10 {prompt = \
		"if output exceeds this, replace with fillval"}
real	fillval = 1.e10 {prompt = "fill value in case result is too large"}
int	norm1 = 31 {prompt = "X size of box for normalizing"}
int	norm2 = 31 {prompt = "Y size of box for normalizing"}
bool	verbose = yes {prompt = "print names of tasks during operation?"}
string	unitmpl = "focutility$tmplrsdpunix.hhh" {prompt = \
		"UNI template header"}
string	mode = "al"

begin
	string	temp		# name of intermediate image
	string	scr1, scr2	# equals either input or temp
	string	in1, in2, intab, outuni
	int	wid1, wid2	# X & Y widths of box for median filter
	int	xrep, yrep	# arguments for blkrep

	if ( ! defpac ("focutility") ) {
	    print ("The focutility package must be loaded.")
	    bye()
	}

	if ( ! defpac ("images") ) {
	    print ("The images package must be loaded.")
	    bye()
	}

	# Get the parameters into local variables.  This is the only
	# place where there is any prompting for parameters.
	in1 = input
	in2 = input2
	intab = intable
	outuni = output

	# The intermediate steps will use this image name.
	temp = mktemp ("flall") // ".hhh"

	# Count how many images there are in the input list.
	flcount (in1)

	# Run imcombine if the input list contains more than one image.
	# scr1 is the name of the image to use in the next step.
	if (flcount.number > 1) {
	    if (verbose) print ("running imcombine")
	    imcombine (in1,
		temp, sigma="", logfile="STDOUT", option="sum", outtype="",
		expname="", exposure=no, scale=no, offset=no, weight=no,
		lowreject=3., highreject=3., blank=0.)
	    scr1 = temp
	} else {
	    scr1 = in1
	}

	# Patch in a corner of an external flat, if one was specified.
	# scr2 is the name of the image to use in the next step.
	if (in2 != "") {
	    if (verbose) print ("running flcorner")
	    flcorner (input1=scr1, input2=in2,
		    output=temp, y1=y1, y2=y2, slope=slope)
	    scr2 = temp
	} else {
	    scr2 = scr1
	}

	# Trim a border, filling in with adjacent pixels.
	if (verbose) print ("running flborder")
	flborder (scr2, temp, left_b=left_b, right_b=right_b,
		bottom_b=bottom_b, top_b=top_b)

	# Interpolate along vertical strips according to a mask table,
	# if one was specified.
	if (intab != "") {
	    if (verbose) print ("running flvmask")
	    flvmask (temp, intab, temp)
	} else {
	    ;
	}

	# Shall we median filter the image?
	if (medfilt) {
	    # Filtering is done on a zoom-format image, so the X scale
	    # is half the Y scale.
	    wid1 = width
	    wid2 = 2 * wid1 + 1

	    # Median filter.
	    if (verbose) print ("running median")
	    median (temp, temp, wid1, wid2, boundary="nearest", constant=0.)
	} else {
	    ;
	}

	# Shall we smooth the image?
	if (sm_option != "none") {
	    # Average to make a 512 x 512 image to save time during smoothing.
	    if (verbose) print ("running blkavg")
	    blkavg (temp, temp, 1, 2, 1, 1, 1, 1, 1)
	    xrep = 2			# we need this for blkrep
	    yrep = 2
	} else {
	    xrep = 2
	    yrep = 1
	}

	if (sm_option == "gauss") {
	    # Gaussian filter.
	    if (verbose) print ("running gauss")
	    gauss (temp, temp, sigma, ratio=1.0, theta=0., nsigma=nsigma,
		boundary="nearest", constant=0.)
	} else if (sm_option == "boxcar") {
	    # Boxcar filter.
	    if (verbose) print ("running boxcar")
	    boxcar (temp, temp, box, box, boundary="nearest", constant=0.)
	} else {
	    ;
	}

	# Expand to 1024 x 1024.
	if (verbose) print ("running blkrep")
	blkrep (temp, temp, xrep, yrep, 1, 1, 1, 1, 1)

	# Take the reciprocal of the image, and convert to RSDP UNI format.
	if (verbose) print ("running rsdpunix")
	rsdpunix (temp, outuni,
		threshold=threshold, fillval=fillval, template=unitmpl)

	# Done with intermediate image.
	imdelete (temp)

	# Normalize the result to unit average within the central norm1 x norm2.
	if (verbose) print ("running flnorm")
	flnorm (outuni, outuni, box1=norm1, box2=norm2)
end
