include <imhdr.h>
include <error.h>
include "fflincorr.h"

define	SZ_KEYVAL	19	# size of a short string (e.g keyword value)
define	VERSION		"1.0"	# version number of fflincorr

# fflincorr -- correct FOC images for saturation
# This task is an implementation of Robert Jedrzejewski's algorithm
# for correcting flat-field images for saturation.  See the Instrument
# Science Report FOC-062 for details.
#
# Phil Hodge,  7-Jul-1992  Task created.

procedure fflincorr()

pointer inlist			# input list pointer
pointer outlist			# output list pointer
pointer a_list			# list pointer for file of A values
int	boundary		# type of boundary extension
real	b_const			# constant boundary extension value
int	box			# size of box for smoothing
int	xbox			# X size of box for zoomed images
bool	verbose			# print file names?
#--
pointer sp
pointer input			# scratch for an image name from inlist
pointer output			# scratch for an image name from outlist
pointer a_file			# scratch for name of file of A values
pointer scratch			# scratch for temporary image name
pointer history			# scratch for string to add to header
pointer outdir			# scratch for output directory name
pointer iim, oim		# imhdr pointers for input & output images
pointer ap			# constants for current image
char	fflcorr[SZ_KEYVAL]	# has linearity correction already been done?
char	optcrly[SZ_KEYVAL]	# F48 or F96 relay
int	nin			# number of names in input list
int	nout			# number of names in output list
int	nafiles			# number of names in a_file list
int	junk
bool	is_dir			# true if output name is a directory
bool	inplace			# true if output image same as input
bool	bad_relay		# relay mismatch between image and a_file?
pointer imtopenp(), immap()
int	imtgetim(), imtlen()
int	isdirectory()
int	imaccf()
bool	streq()
real	clgetr()
int	clgwrd(), clgeti()
bool	clgetb()

begin
	call smark (sp)
	call salloc (input, SZ_FNAME, TY_CHAR)
	call salloc (output, SZ_PATHNAME, TY_CHAR)
	call salloc (a_file, SZ_FNAME, TY_CHAR)
	call salloc (scratch, SZ_FNAME, TY_CHAR)
	call salloc (history, SZ_FNAME, TY_CHAR)
	call salloc (outdir, SZ_PATHNAME, TY_CHAR)

	inlist = imtopenp ("input")
	outlist = imtopenp ("output")
	a_list = imtopenp ("a_file")

	# The X size of the smoothing box with be either xbox or box,
	# depending on zoom.
	box = clgeti ("box")
	xbox = clgeti ("xbox")
	boundary = clgwrd ("boundary", Memc[outdir], SZ_PATHNAME,
	    ",constant,nearest,reflect,wrap,")
	if (boundary < 1)
	    call error (1, "invalid boundary extension")
	if (boundary == 1)
	    b_const = clgetr ("b_const")
	else
	    b_const = 0.
	verbose = clgetb ("verbose")

	nin = imtlen (inlist)
	nout = imtlen (outlist)
	nafiles = imtlen (a_list)
	if (nin <= 0)
	    call error (1, "no input specified")
	if (nout <= 0)
	    call error (1, "no output specified")

	is_dir = false				# initial value
	if (nout == 1) {
	    # Get the output name.
	    junk = imtgetim (outlist, Memc[output], SZ_PATHNAME)
	    # Check whether it's a directory.
	    if (isdirectory (Memc[output], Memc[outdir], SZ_PATHNAME) > 0)
		is_dir = true
	} else if (nin != nout) {
	    call error (1, "input & output lists not the same length")
	}

	# There can be just one a_file to apply to all input, or there
	# can be one a_file for every input file.
	ap = NULL
	if (nafiles == 1) {
	    junk = imtgetim (a_list, Memc[a_file], SZ_FNAME)
	    call lin_get_a (Memc[a_file], ap)		# read values
	} else if (nafiles != nin) {
	    call error (1, "input & a_file lists not the same length")
	}

	# Do for each image in the input list.
	while (imtgetim (inlist, Memc[input], SZ_FNAME) != EOF) {

	    # Either append the input name to the output directory,
	    # or get the corresponding output image name.
	    if (is_dir)
		call lin_append (Memc[input], Memc[outdir],
			Memc[output], SZ_PATHNAME)
	    else if (nout > 1)			# else we already have it
		junk = imtgetim (outlist, Memc[output], SZ_PATHNAME)

	    # Get the name of the file of constants (a values).
	    if (nafiles > 1)
		junk = imtgetim (a_list, Memc[a_file], SZ_FNAME)

	    if (verbose) {
		call printf ("%s --> %s using %s\n")
		    call pargstr (Memc[input])
		    call pargstr (Memc[output])
		    call pargstr (Memc[a_file])
		call flush (STDOUT)
	    }

	    # Read the arrays of constants.
	    if (nafiles > 1) {
		if (ap != NULL)
		    call mfree (ap, TY_STRUCT)
		call lin_get_a (Memc[a_file], ap)
	    }

	    iim = immap (Memc[input], READ_ONLY, NULL)

	    # Make sure this image has not already been corrected.
	    if (imaccf (iim, "fflcorr") == YES) {
		call imgstr (iim, "fflcorr", fflcorr, SZ_KEYVAL)
		call strupr (fflcorr)
		if (streq (fflcorr, "COMPLETE")) {
		    call eprintf ("`%s' has already been corrected\n")
			call pargstr (Memc[input])
		    next
		}
	    }

	    # Which relay is this?
	    if (A_RELAY(ap) > 0) {
		call imgstr (iim, "optcrly", optcrly, SZ_KEYVAL)
		call strupr (optcrly)
		bad_relay = false
		if (streq (optcrly, "F48")) {
		    if (A_RELAY(ap) != 48)
			bad_relay = true
		} else {
		    if (A_RELAY(ap) != 96)
			bad_relay = true
		}
		if (bad_relay) {
		    call eprintf ("relays not the same in %s & %s\n")
			call pargstr (Memc[input])
			call pargstr (Memc[a_file])
		    next
		}
	    }

	    # Open the output image.  If the input and output names are the
	    # same, write to a scratch image, and later overwrite the input.
	    if (streq (Memc[input], Memc[output])) {
		inplace = true
		call mktemp ("lin", Memc[scratch], SZ_FNAME)
		oim = immap (Memc[scratch], NEW_COPY, iim)
	    } else {
		inplace = false
		oim = immap (Memc[output], NEW_COPY, iim)
	    }

	    # Make the correction.
	    iferr (call xfflincorr (iim, oim, ap,
			xbox, box, boundary, b_const)) {
		call eprintf ("Skip image `%s' because:\n")
		    call pargstr (Memc[input])
		call erract (EA_WARN)
		call imunmap (oim)
		call imunmap (iim)
		if (inplace)
		    call imdelete (Memc[scratch])
		else
		    call imdelete (Memc[output])
		next
	    }

	    # Set FFLCORR keyword, and add history info.
	    call imastr (oim, "fflcorr", "COMPLETE")
	    call strcpy (
		"Corrected for flat-field saturation by fflincorr version ",
			Memc[history], SZ_FNAME)
	    call strcat (VERSION, Memc[history], SZ_FNAME)
	    call imputh (oim, "history", Memc[history])
	    call sprintf (Memc[history], SZ_FNAME, "input %s, using %s")
		call pargstr (Memc[input])
		call pargstr (Memc[a_file])
	    call imputh (oim, "history", Memc[history])

	    # Close files.
	    call imunmap (oim)
	    call imunmap (iim)
	    if (inplace) {
		call imdelete (Memc[input])
		call imrename (Memc[scratch], Memc[output])
	    }
	}

	if (ap != NULL)
	    call mfree (ap, TY_STRUCT)
	call imtclose (outlist)
	call imtclose (inlist)
	call sfree (sp)
end

# lin_append -- append image name to directory
# This is based on t_imcopy.x.
# The input image name may contain a directory prefix and/or section.
# This name without either directory or section is appended to the
# output directory name to form the name of the output image.

procedure lin_append (input, outdir, output, maxch)

char	input[ARB]		# i: name of input image
char	outdir[ARB]		# i: name of output directory
char	output[ARB]		# o: name of output image
int	maxch			# i: size of output string
#--
pointer sp
pointer image			# input name without image section
pointer scr			# ignored
int	root_len		# length of logical directory prefix
int	fnldir()

begin
	call smark (sp)
	call salloc (image, SZ_PATHNAME, TY_CHAR)
	call salloc (scr, SZ_PATHNAME, TY_CHAR)

	# Strip the image section first because fnldir recognizes it
	# as part of a directory.

	call get_root (input, Memc[image], SZ_PATHNAME)
	root_len = fnldir (Memc[image], Memc[scr], SZ_PATHNAME)

	# This is just the directory name ...
	call strcpy (outdir, output, maxch)

	# ... now append the input image name without its directory
	# or image section.
	call strcat (Memc[image+root_len], output, maxch)

	call sfree (sp)
end

# xfflincorr -- apply linearity correction

procedure xfflincorr (iim, oim, ap, xbox, nyk, boundary, b_const)

pointer iim			# i: imhdr pointer for input image
pointer oim			# i: imhdr pointer for output image
pointer ap			# i: struct containing A values
int	xbox			# i: X size of box for zoomed images
int	nyk			# i: Y size of the smoothing kernel
int	boundary		# i: type of boundary extnsion
real	b_const			# i: value for constant boundary extension
#--
pointer sp
pointer smooth			# scratch for smoothed image
pointer ix, ox			# pointers to data for input & output
int	nxk			# X size of the smoothing kernel (=nyk or xbox)
int	n1, n2			# lengths of first & second axes
int	j
pointer imgl2r(), impl2r()

# keyword values and related info:
real	exptime
int	sampoff			# value of sampoff from header, but "flipped"
int	lineoff			# value of lineoff from header
int	zoom			# zoom factor:  1 or 2
bool	dezoomed		# true if data have been dezoomed
bool	subset			# true if image is just a subset of original

errchk	lin_get_head, lin_boxcar, lin_factor

begin
	call smark (sp)

	# Assign a value to one pixel to create the data file
	# so we can close the image in case of error.
	ox = impl2r (oim, 1)
	Memr[ox] = 0.

	n1 = IM_LEN(iim,1)
	n2 = IM_LEN(iim,2)

	# Get header parameters.
	call lin_get_head (iim, exptime,
		sampoff, lineoff, zoom, dezoomed, subset)
	if (subset && (A_NX(ap) > 1 || A_NY(ap) > 1))
	    call error (1, "can't correct just a subset of original image")

	# Set the X length of the smoothing kernel, depending on the zoom.
	if (dezoomed || zoom < 2)
	    nxk = nyk
	else
	    nxk = xbox

	# Allocate space for a copy of the input image, do a boxcar average,
	# and divide by exposure time to get rho, the observed count rate.
	call salloc (smooth, n1 * n2, TY_REAL)

	call lin_boxcar (iim, Memr[smooth], n1, n2, nxk, nyk,
		    boundary, b_const)
	call adivkr (Memr[smooth], exptime, Memr[smooth], n1 * n2)

	# From the observed count rate, compute the ITF correction factor
	# r/rho, which replaces the smoothed image value.
	# This factor is adjusted for image size and pixel format.
	call lin_factor (ap, Memr[smooth], n1, n2,
		sampoff, lineoff, zoom, dezoomed,
		A_VAL(ap,1,1), A_NX(ap), A_NY(ap))

	# Apply the correction factor to each pixel in the image.
	do j = 1, n2 {
	    ix = imgl2r (iim, j)
	    ox = impl2r (oim, j)
	    call amulr (Memr[smooth+(j-1)*n1], Memr[ix], Memr[ox], n1)
	}

	call sfree (sp)
end

# lin_get_head -- get header parameters
# The dezoomed parameter will be true only if the image has actually been
# dezoomed; dezoomed will be false if the image was not in zoom mode in
# the first place.

procedure lin_get_head (iim, exptime,
		sampoff, lineoff, zoom, dezoomed, subset)

pointer iim			# i: imhdr pointer for input image
real	exptime			# o: exposure time in seconds
int	sampoff			# o: value of sampoff from header, but "flipped"
int	lineoff			# o: value of lineoff from header
int	zoom			# o: zoom factor:  1 or 2
bool	dezoomed		# o: true if data have been dezoomed
bool	subset			# o: true if image is just a subset of original
#--
char	pxformt[SZ_KEYVAL], pxlcorr[SZ_KEYVAL]
int	nsamps			# samppln, or 2*samppln if dezoomed
int	samppln, linepfm
real	imgetr()
int	imgeti()
bool	streq()
errchk	imgeti, imgetr, imgstr

begin
	exptime = imgetr (iim, "exptime")
	if (exptime < 0.) {
	    call eprintf ("exptime is negative -- will change sign\n")
	    exptime = -exptime
	} else if (exptime == 0.) {
	    call error (1, "exptime is zero")
	}

	samppln = imgeti (iim, "samppln")
	linepfm = imgeti (iim, "linepfm")
	nsamps = samppln
	sampoff = nint (imgetr (iim, "sampoff"))
	lineoff = nint (imgetr (iim, "lineoff"))

	call imgstr (iim, "pxformt", pxformt, SZ_KEYVAL)
	call strupr (pxformt)
	if (streq (pxformt, "ZOOM")) {
	    zoom = 2
	    call imgstr (iim, "pxlcorr", pxlcorr, SZ_KEYVAL)
	    call strupr (pxlcorr)
	    if (streq (pxlcorr, "COMPLETE")) {
		dezoomed = true
		nsamps = samppln * 2		# to compare with naxis1
	    } else {
		dezoomed = false
	    }
	} else {
	    zoom = 1
	    dezoomed = false
	}

	# "flip" sampoff.
	sampoff = 1024 - zoom*samppln - sampoff

	# Check whether we have the complete image or just a subset.
	subset = nsamps != IM_LEN(iim,1) || linepfm != IM_LEN(iim,2)
end
