include "fflincorr.h"

# lin_factor -- compute ITF correction factor
# The input image corr is the smoothed image divided by exposure time
# to give count rate; on output this will be the ITF factor, the ratio
# of actual to observed count rates.
# The value of A_NSAMPS refers to the dezoomed image, and
# sampoff & A_SAMPOFF are "flipped".
#
# Phil Hodge,  7-Jul-1992  Subroutine created.

procedure lin_factor (ap, corr, n1, n2,
		sampoff, lineoff, zoom, dezoomed,
		a_array, nx, ny)

pointer ap		# i: pointer to structure containing A values
real	corr[n1,n2]	# io: input smoothed image; output correction factor
int	n1, n2		# i: size of image
int	sampoff		# i: value of sampoff from header, but "flipped"
int	lineoff		# i: value of lineoff from header
int	zoom		# i: zoom factor, 1 or 2
bool	dezoomed	# i: true if data have been dezoomed
real	a_array[nx,ny]	# i: array of A values
int	nx, ny		# i: size of a_array
#--
real	factor		# to correct for format
real	a_value		# value of A
real	rho		# observed count rate
real	v		# argument to log function
int	a_zoom		# 1 or 2, for comparing n1 with A_NSAMPS
int	i, j

# for interpolating:
real	px, qx		# fraction of distance from endpoint of interval
real	py, qy		# same for y
real	x_off, y_off	# offset of image from A array
real	x_step, y_step	# 1. / (pixels per step in A array)
real	ax, ay		# location in a_array corresponding to corr[i,j]
int	iax, iay	# integer parts of ax, ay
int	iax1, iay1	# iax+1, iay+1
int	last_iax	# previous value of iax
real	left		# interpolated point on left side of box
real	right		# interpolated point on right side of box

begin
	# If the table of A values and the current image are for different
	# formats, we should multiply the A values by this factor to
	# compensate for the difference.  The zoom factors will be 1 or 2.
	factor = A_ZOOM(ap) * A_NSAMPS(ap) * A_NLINES(ap) / (zoom * n1 * n2)

	# If the input image has been dezoomed, n1 must be no larger than
	# A_NSAMPS; if it has not been dezoomed, compare 2*n1 with A_NSAMPS.
	if (zoom == 2 && !dezoomed)
	    a_zoom = 2
	else
	    a_zoom = 1

	# Only one A value for entire image?
	if (nx == 1 && ny == 1) {

	    a_value = A_VAL(ap,1,1)

	    # Relay not specified means we were given an explicit number,
	    # rather than a file of A values, so we should use that number
	    # directly, without applying the format factor.
	    if (A_RELAY(ap) > 0)
		a_value = a_value * factor

	    # Compute the ITF correction.
	    do j = 1, n2 {
		do i = 1, n1 {

		    rho = corr[i,j]
		    if (rho <= 0.) {
			corr[i,j] = 1.
		    } else {
			v = 1. - rho / a_value
			if (v <= 0.)
			    corr[i,j] = MAX_ITF
			else
			    corr[i,j] = -a_value * log (v) / rho
		    }
		}
	   }

	} else {			# we have an array of A values

	    # Check size and offset of image with respect to A array.
	    if (n1*a_zoom > A_NSAMPS(ap) || n2 > A_NLINES(ap))
		call error (1, "image is larger than table of A values")

	    if (sampoff < A_SAMPOFF(ap) ||
		lineoff < A_LINEOFF(ap) ||
		sampoff + n1*a_zoom > A_SAMPOFF(ap) + A_NSAMPS(ap) ||
		lineoff + n2        > A_LINEOFF(ap) + A_NLINES(ap))
		call error (1, "image not contained within table of A values")

	    # Given pixel coordinates X & Y in the image, the corresponding
	    # point (not generally an integer) in the array of A values is:
	    #   ax = (X-1 + x_off) * x_step + 0.5
	    #   ay = (Y-1 + y_off) * y_step + 0.5
	    x_off = sampoff - A_SAMPOFF(ap)
	    y_off = lineoff - A_LINEOFF(ap)

	    x_step = real (A_NX(ap)) / real (A_NSAMPS(ap)) * a_zoom
	    y_step = real (A_NY(ap)) / real (A_NLINES(ap))

	    # Default values, appropriate for A_NY = 1.
	    iay = 1
	    iay1 = 1
	    py = 0.
	    qy = 1.
	    # Compute the ITF factor for each pixel.
	    do j = 1, n2 {			# do for each line

		if (A_NY(ap) > 1) {
		    ay = (j-1 + y_off) * y_step + 0.5
		    iay = max (ay, 1.)
		    iay1 = min (iay + 1, A_NY(ap))
		    iay = iay1 - 1
		    py = ay - iay			# fractional part
		    qy = 1. - py
		}

		iax = 1				# default values for A_NX = 1
		iax1 = 1
		px = 0.
		qx = 1.
		last_iax = -100				# initial value
		do i = 1, n1 {			# do for each pixel on line

		    # Find the location (ax,ay) of [i,j] in the A array.
		    if (A_NX(ap) > 1) {
			ax = (i-1 + x_off) * x_step + 0.5
			iax = max (ax, 1.)
			iax1 = min (iax + 1, A_NX(ap))
			iax = iax1 - 1
			px = ax - iax
			qx = 1. - px
		    }

		    if (iax != last_iax) {
			# New box in A array; interpolate in Y at the
			# left and right sides of the box.
			left = qy * a_array(iax,iay) + py * a_array(iax,iay1)
			right = qy * a_array(iax1,iay) + py * a_array(iax1,iay1)
			last_iax = iax
		    }

		    # Now interpolate in X to get the A value for [i,j].
		    a_value = (qx * left + px * right) * factor

		    # Compute the ITF correction and overwrite corr[i,j].
		    rho = corr[i,j]
		    if (rho <= 0.) {
			corr[i,j] = 1.
		    } else {
			v = 1. - rho / a_value
			if (v <= 0.)
			    corr[i,j] = MAX_ITF
			else
			    corr[i,j] = -a_value * log (v) / rho
		    }
		}
	    }
	}
end
