#  FILTERPHOT -- Compute instrumental or standard magnitudes and colors 
#
#  Description:
#  ------------
#  This procedure computes the magnitudes and the colors of objects from their
#  intensities and transforms these magnitudes and colors to standard 
#  photometric systems.
#
#  The user can select one of the following operations:
#  1) Instrumental magnitudes can be computed from intensity measurements.
#  The instrumental magnitude IM is calculated from the instrumental intensity I#  via the formula
#
#		IM = -2.5 * LOG (I) + IM0
# 
#  where IM0 is the magnitude zero point constant and LOG represents the common
#  logarithm function.
#
#  2) Instrumental colors can be computed from two intensity measurements.  The
#  instrumental color IC is calculated from the instrumental intensity
#  measurements I1 and I2 obtained through different filters via the formula
#
#		IC = -2.5 * LOG (I1 / I2) + IC0
# 
#  where IC0 is the color zero point constant and LOG represents the common
#  logarithm function.
#
#  3) Instrumental magnitudes can be transformed to a standard photometric
#  system.  The standardized magnitude SM is calculated from the instrumental
#  magnitude IM and certain apropriate (standard) color C' via the formula
# 
#		SM = a + b * IM + c * C'
#
#  where a, b and c are magnitude transformation constants.
#
#  4) Instrumental colors can be transformed to a standard photometric
#  system.  The standardized color SC is calculated from the instrumental
#  colors IC and (maybe) a certain apropriate color C' via the formula
#
#		SC = r + s * IC + t * C'
#
#  where r, s and t color transformation constants.
#
#  This procedure assumes that the input measurements are calibrated
#  count rates.
#  Sometimes, the transformation of magnitudes or colors to a standard
#  photometric system requires a formula more complicated than those which have
#  been implemented.  Multiple invocations of this procedure with the command
#  language can be used to effect the transformation.  
#
#  Date		Author			Description
#  ----		------			-----------
#  11-Apr-1986	D. Ball		        Original module
#  09-May-1990  J.-C. Hsu		rewrite in SPP
#------------------------------------------------------------------------------

procedure filterphot ()

int	option			# operation to be performed
pointer	fin1, finmask1, fin2, finmask2, fout, foutmask	
				# file template pointers
int	nfin1, nfin2		# numbers of input files 
bool	mask1, mask2		# are there input mask files?
real	coeff[3]		# transformation coefficients
real	datafill		# fill value for the invalid pixels in the 
				# output data file
#==============================================================================
begin

	# announce start of the task
	call printf ("*** FILTERPHOT - Version 1.0 ***\n")

	# get input parameters
	call fphot_get (fin1, finmask1, fin2, finmask2, fout, foutmask, 
				nfin1, nfin2, mask1, mask2, option, coeff, 
			  	datafill)

	# do the filter photometry calculation
	call fphot_do (fin1, finmask1, fin2, finmask2, fout, foutmask, 
				nfin1, nfin2, mask1, mask2, option, coeff, 
			  	datafill)

	# close file templates
	call imtclose (fin1)
	call imtclose (finmask1)
	call imtclose (fin2)
	call imtclose (finmask2)
	call imtclose (fout)
	call imtclose (foutmask)

	# announce completion of task
	call printf ("filter photometry reduction task completed with no error\n")
end
