#---------------------------------------------------------------------------
# pha_diode -- Produce graphs and statistcs about a specific diode(s)
#---------------------------------------------------------------------------
procedure pha_diode(diodes)

file    diodes          {prompt="File containing list of diodes to examine"}

file    intphd          {"",prompt="Integral pulse height distribution image"}
file    difphd          {"",prompt="Differential pulse height distribution image"}
file    fitphd          {"",prompt="Fitted pulse height distribution image"}
file    table           {"",prompt="Threshold statistics table"}
real    t0              {0,prompt="Initial threshold level in input"}
real    delt            {0,prompt="Step size in theshold levels in input"}
string  device          {"stdgraph",prompt="Graphics output device"}
real    left            {0.,min=0.,max=1.,prompt="Left edge of viewport"}
real    right           {1.,min=0.,max=1.,prompt="Right edge of viewport"}
real    bottom          {0.,min=0.,max=1.,prompt="Bottom edge of viewport"}
real    top             {1.,min=0.,max=1.,prompt="Top edge of viewport"}
bool    verbose         {no,prompt="Write status messages?"}

string  Version         {"11Dec92",prompt="Date of Installation"}
struct  *diode_list     {mode="h"}

begin
        # Declarations
        real    center          # Where the gaussian center is.
        real    hv              # High voltage
        real    x, y            # Generic.

        int     detector        # Detector.
        int     diode           # Current diode number working on.
        int     i               # Generic.
        int     thresh          # Threshold value.
        
        file    igi_script      # The igi script to execute.
        file    thresholds      # List of thresholds.
        file    tmp_table_1     # Sorted table on diodes.
        
        string  pdiodes         # The diodes to print up.
        string  rootname        # Rootname of observations.
        string  ts1             # Generic

        # Set temporary file names
        igi_script = mktemp ("tmp$pha_igi")
        thresholds = mktemp ("tmp$pha_thresh")
        tmp_table_1 = mktemp ("tmp$pha_table")//".tab"
        
        # Make sure the correct packages are loaded.
        ts1 = ""
        if (!defpac ("stplot"))
            ts1 = ts1//"stplot "
        if (!defpac ("ttools"))
            ts1 = ts1//"ttools "
	if (!defpac ("images"))
	    ts1 = ts1//"images "
	if (!defpac ("z_calib"))
	    ts1 = ts1//"z_calib "
        if (strlen (ts1) > 0)
            error (1, "Please load packages: "//ts1)

        # Retrieve the basic parameters.
        pdiodes = diodes

        # Take values from the phaz task if not defined here.
        if (strlen (intphd) == 0) {
            ts1 = phaz.intphd.p_mode
            phaz.intphd.p_mode = "h"
            intphd = phaz.intphd
            phaz.intphd.p_mode = ts1
        }
        if (strlen (difphd) == 0) {
            ts1 = phaz.difphd.p_mode
            phaz.difphd.p_mode = "h"
            difphd = phaz.difphd
            phaz.difphd.p_mode = ts1
        }
        if (strlen (fitphd) == 0) {
            ts1 = phaz.fitphd.p_mode
            phaz.fitphd.p_mode = "h"
            fitphd = phaz.fitphd
            phaz.fitphd.p_mode = ts1
        }
        if (strlen (table) == 0) {
            ts1 = phaz.table.p_mode
            phaz.table.p_mode = "h"
            table = phaz.table
            phaz.table.p_mode = ts1
        }
        if (t0 == 0.0) {
            ts1 = phaz.t0.p_mode
            phaz.t0.p_mode = "h"
            t0 = phaz.t0
            phaz.t0.p_mode = ts1
        }
        if (delt == 0.0) {
            ts1 = phaz.delt.p_mode
            phaz.delt.p_mode = "h"
            delt = phaz.delt
            phaz.delt.p_mode = ts1
        }

        # Create a data file with the list of threshold values.
        delete (thresholds, go_ahead=yes, verify=no, >& "dev$null")
        keypar (intphd, "i_naxis1")
        x = real (keypar.value)
        for ( i = 0; i < x; i = i + 1 )
            print ((i*delt)+t0, >> thresholds)

        # Sort the table on diodes.
        delete (tmp_table_1, go_ahead=yes, verify=no, >& "dev$null")
        tcopy (table, tmp_table_1, verbose=no)
        tsort (tmp_table_1, columns="DIODE", ascend=yes, casesens=yes)

        # Get various keyword parameters.
        keypar (intphd, "rootname")
        rootname = keypar.value
        keypar (intphd, "detector")
        detector = int (keypar.value)
        if (verbose)
            print ("pha_diode: Rootname "//rootname//" Detector "//detector)
        keypar (intphd, "HV")
        hv = real (keypar.value)
        
        # For each diode listed, produce the information.
        diode_list = pdiodes
        while (fscan (diode_list, ts1) != EOF) {
            diode = int (ts1)
            if (verbose)
                print ("pha_diode: Diode "//diode)
            
            # Prepare the igi script.
            delete (igi_script, go_ahead=yes, verify=no, >& "dev$null")

            # First draw the integrated pha for the diode.
            print ("location 0.5 0.95 0.55 0.95", >> igi_script)
            print ("data "//thresholds//"; xcolumn 1", >> igi_script)
            print ("ysection "//intphd//"[*,"//diode//"]", >> igi_script)
            print ("lweight 1.5; limits; box; connect", >> igi_script)

            # Draw the derivative and overplot the fit.  Also show the
            # selected threshold.
            ts1 = difphd//"[*,"//diode//"]"
            print ("location 0.5 0.95 .05 .45", >> igi_script)
            print ("ysection "//ts1, >> igi_script)
            print ("ptype 1 3; limits ; box ; points", >> igi_script)

            print ("ysection "//fitphd//"[*,"//diode//"]", >> igi_script)
            print ("connect", >> igi_script)

            minmax (ts1, force=yes, update=no, verbose=no)
            tabpar (tmp_table_1, "THRESH", diode)
            thresh = int (tabpar.value)
            print ("ltype 2", >> igi_script)
            print ("move "//thresh//" "//minmax.minval, >> igi_script)
            print ("draw "//thresh//" "//minmax.maxval, >> igi_script)

            # Now write some text with the pertinent information.
            print ("location 0 1 0 1; limits 0 1 0 1; justify 5; move .25 .9",
                   >> igi_script)
            print ("lweight 1.5 ; expand 1.5" , >> igi_script)
            print ("label ID = "//rootname, >> igi_script)
            print ("move .25 .85", >> igi_script)
            print ("label Detector "//detector//" Diode "//diode,
                   >> igi_script)
            print ("expand 1.0 ; justify 4", >> igi_script)
            print ("move .25 .80 ; label amp = ", >> igi_script)
            print ("move .25 .77 ; label chan = ", >> igi_script)
            print ("move .25 .74 ; label data-item = ", >> igi_script)
            print ("move .25 .71 ; label signal peak = ", >> igi_script)
            print ("move .25 .68 ; label rms noise = ", >> igi_script)
            print ("move .25 .65 ; label sig/noise = ", >> igi_script)
            print ("move .25 .62 ; label FWHM (kev) = ", >> igi_script)
            print ("move .25 .59 ; label thresh select = ", >> igi_script)
            print ("move .25 .56 ; label thresh counts = ", >> igi_script)
            print ("move .25 .53 ; label thresh cps = ", >> igi_script)

            print ("justify 6", >> igi_script)
            tabpar (tmp_table_1, "AMP", diode)
            print ("move .25 .80 ; label "//tabpar.value, >> igi_script)
            tabpar (tmp_table_1, "CHAN", diode)
            print ("move .25 .77 ; label "//tabpar.value, >> igi_script)
            tabpar (tmp_table_1, "DATAITEM", diode)
            print ("move .25 .74 ; label "//tabpar.value, >> igi_script)
            tabpar (tmp_table_1, "CENTER", diode)
            center = real (tabpar.value)
            print ("move .25 .71 ; label "//tabpar.value, >> igi_script)
            tabpar (tmp_table_1, "SIGMA", diode)
            y = real (tabpar.value)
            print ("move .25 .68 ; label "//tabpar.value, >> igi_script)
            if (y == 0.)
                print ("move .25 .65 ; label UNDEFINED", >> igi_script)
            else {
                x = center / y
                printf ("move .25 .65 ; label %7.3f\n", x, >> igi_script)
            }
            if (center == 0.)
                print ("move .25 .62 ; label UNDEFINED", >> igi_script)
            else {
                tabpar (tmp_table_1, "FWHM", diode)
                x = real (tabpar.value) * hv / center
                printf ("move .25 .62 ; label %4.2f\n", x, >> igi_script)
            }
            print ("move .25 .59 ; label "//thresh, >> igi_script)
            tabpar (tmp_table_1, "COUNTS", diode)
            i = int (tabpar.value)
            print ("move .25 .56 ; label "//i, >> igi_script)
            keypar (intphd, "EXPOSURE")
            x = real (keypar.value)
            x = i / x
            printf ("move .25 .53 ; label %7.1f\n", x, >> igi_script)
            
            # Do the script.
            print ("end", >> igi_script)
            flprcache
            igi (initcmd="", device=device, metacode="", append=no,
                 debug=no,cursor="", < igi_script, >& "dev$null")
        }

        # That's all folks.
        if (verbose)
            print ("pha_diode: That's all folks")
        gflush
        delete (igi_script, go_ahead=yes, verify=no, >& "dev$null")
        delete (thresholds, go_ahead=yes, verify=no, >& "dev$null")
        delete (tmp_table_1, go_ahead=yes, verify=no, >& "dev$null")

end
#---------------------------------------------------------------------------
# End of pha_diode
#---------------------------------------------------------------------------
